@@ -62,9 +62,18 @@ var PointerlockFsWarning = {
       this._element = document.getElementById(elementId);
       // Setup event listeners
       this._element.addEventListener("transitionend", this);
+      this._element.addEventListener("transitioncancel", this);
       window.addEventListener("mousemove", this, true);
+      // If the user explicitly disables the prompt, there's no need to detect
+      // activation.
+      if (timeout > 0) {
+        window.addEventListener("activate", this);
+        window.addEventListener("deactivate", this);
+      }
       // The timeout to hide the warning box after a while.
       this._timeoutHide = new this.Timeout(() => {
+        window.removeEventListener("activate", this);
+        window.removeEventListener("deactivate", this);
         this._state = "hidden";
       }, timeout);
       // The timeout to show the warning box when the pointer is at the top
@@ -90,13 +99,14 @@ var PointerlockFsWarning = {
       ".pointerlockfswarning-domain-text"
     );
     if (!host) {
-      textElem.setAttribute("hidden", true);
+      textElem.hidden = true;
     } else {
       textElem.removeAttribute("hidden");
       // Document's principal's URI has a host. Display a warning including it.
-      let utils = {};
-      ChromeUtils.import("resource://gre/modules/DownloadUtils.jsm", utils);
-      let displayHost = utils.DownloadUtils.getURIHost(uri.spec)[0];
+      let { DownloadUtils } = ChromeUtils.importESModule(
+        "resource://gre/modules/DownloadUtils.sys.mjs"
+      );
+      let displayHost = DownloadUtils.getURIHost(uri.spec)[0];
       let l10nString = {
         "fullscreen-warning": "fullscreen-warning-domain",
         "pointerlock-warning": "pointerlock-warning-domain",
@@ -115,15 +125,23 @@ var PointerlockFsWarning = {
       return;
     }
 
-    // Explicitly set the last state to hidden to avoid the warning
-    // box being hidden immediately because of mousemove.
-    this._state = "onscreen";
-    this._lastState = "hidden";
-    this._timeoutHide.start();
+    if (Services.focus.activeWindow == window) {
+      this._state = "onscreen";
+      this._timeoutHide.start();
+    }
   },
 
-  close() {
-    if (!this._element) {
+  /**
+   * Close the full screen or pointerlock warning.
+   * @param {('fullscreen-warning'|'pointerlock-warning')} elementId - Id of the
+   * warning element to close. If the id does not match the currently shown
+   * warning this is a no-op.
+   */
+  close(elementId) {
+    if (!elementId) {
+      throw new Error("Must pass id of warning element to close");
+    }
+    if (!this._element || this._element.id != elementId) {
       return;
     }
     // Cancel any pending timeout
@@ -131,14 +149,17 @@ var PointerlockFsWarning = {
     this._timeoutShow.cancel();
     // Reset state of the warning box
     this._state = "hidden";
+    this._doHide();
     // Reset state of the text so we don't persist or retranslate it.
     this._element
       .querySelector(".pointerlockfswarning-domain-text")
       .removeAttribute("data-l10n-id");
-    this._element.setAttribute("hidden", true);
     // Remove all event listeners
     this._element.removeEventListener("transitionend", this);
+    this._element.removeEventListener("transitioncancel", this);
     window.removeEventListener("mousemove", this, true);
+    window.removeEventListener("activate", this);
+    window.removeEventListener("deactivate", this);
     // Clear fields
     this._element = null;
     this._timeoutHide = null;
@@ -165,6 +186,14 @@ var PointerlockFsWarning = {
     }
     return "hiding";
   },
+
+  _doHide() {
+    try {
+      this._element.hidePopover();
+    } catch (e) {}
+    this._element.hidden = true;
+  },
+
   set _state(newState) {
     let currentState = this._state;
     if (currentState == newState) {
@@ -174,24 +203,12 @@ var PointerlockFsWarning = {
       this._lastState = currentState;
       this._element.removeAttribute(currentState);
     }
+    if (currentState == "hidden") {
+      this._element.showPopover();
+    }
+    // hidden is dealt with on transitionend or close(), see _doHide().
     if (newState != "hidden") {
-      if (currentState != "hidden") {
-        this._element.setAttribute(newState, true);
-      } else {
-        // When the previous state is hidden, the display was none,
-        // thus no box was constructed. We need to wait for the new
-        // display value taking effect first, otherwise, there won't
-        // be any transition. Since requestAnimationFrame callback is
-        // generally triggered before any style flush and layout, we
-        // should wait for the second animation frame.
-        requestAnimationFrame(() => {
-          requestAnimationFrame(() => {
-            if (this._element) {
-              this._element.setAttribute(newState, true);
-            }
-          });
-        });
-      }
+      this._element.setAttribute(newState, "");
     }
   },
 
@@ -207,7 +224,7 @@ var PointerlockFsWarning = {
           } else if (this._timeoutShow.delay >= 0) {
             this._timeoutShow.start();
           }
-        } else {
+        } else if (state != "onscreen") {
           let elemRect = this._element.getBoundingClientRect();
           if (state == "hiding" && this._lastState != "hidden") {
             // If we are on the hiding transition, and the pointer
@@ -229,10 +246,24 @@ var PointerlockFsWarning = {
         }
         break;
       }
-      case "transitionend": {
+      case "transitionend":
+      case "transitioncancel": {
         if (this._state == "hiding") {
-          this._element.setAttribute("hidden", true);
+          this._doHide();
         }
+        if (this._state == "onscreen") {
+          window.dispatchEvent(new CustomEvent("FullscreenWarningOnScreen"));
+        }
+        break;
+      }
+      case "activate": {
+        this._state = "onscreen";
+        this._timeoutHide.start();
+        break;
+      }
+      case "deactivate": {
+        this._state = "hidden";
+        this._timeoutHide.cancel();
         break;
       }
     }
@@ -240,12 +271,25 @@ var PointerlockFsWarning = {
 };
 
 var PointerLock = {
+  _isActive: false,
+
+  /**
+   * @returns {boolean} - true if pointer lock is currently active for the
+   * associated window.
+   */
+  get isActive() {
+    return this._isActive;
+  },
+
   entered(originNoSuffix) {
+    this._isActive = true;
+    Services.obs.notifyObservers(null, "pointer-lock-entered");
     PointerlockFsWarning.showPointerLock(originNoSuffix);
   },
 
   exited() {
-    PointerlockFsWarning.close();
+    this._isActive = false;
+    PointerlockFsWarning.close("pointerlock-warning");
   },
 };
 
@@ -257,6 +301,11 @@ var FullScreen = {
       "permissions.fullscreen.allowed"
     );
 
+    let notificationExitButton = document.getElementById(
+      "fullscreen-exit-button"
+    );
+    notificationExitButton.addEventListener("click", this.exitDomFullScreen);
+
     // Called when the Firefox window go into fullscreen.
     addEventListener("fullscreen", this, true);
 
@@ -266,6 +315,7 @@ var FullScreen = {
     // the content.
     addEventListener("willenterfullscreen", this, true);
     addEventListener("willexitfullscreen", this, true);
+    addEventListener("MacFullscreenMenubarRevealUpdate", this, true);
 
     if (window.fullScreen) {
       this.toggle();
@@ -284,6 +334,12 @@ var FullScreen = {
     }
   },
 
+  get fullScreenToggler() {
+    delete this.fullScreenToggler;
+    return (this.fullScreenToggler =
+      document.getElementById("fullscr-toggler"));
+  },
+
   toggle() {
     var enterFS = window.fullScreen;
 
@@ -300,44 +356,36 @@ var FullScreen = {
       // Make sure the menu items are adjusted.
       document.getElementById("enterFullScreenItem").hidden = enterFS;
       document.getElementById("exitFullScreenItem").hidden = !enterFS;
+      this.shiftMacToolbarDown(0);
     }
 
-    if (!this._fullScrToggler) {
-      this._fullScrToggler = document.getElementById("fullscr-toggler");
-      this._fullScrToggler.addEventListener("mouseover", this._expandCallback);
-      this._fullScrToggler.addEventListener("dragenter", this._expandCallback);
-      this._fullScrToggler.addEventListener("touchmove", this._expandCallback, {
-        passive: true,
-      });
-    }
+    let fstoggler = this.fullScreenToggler;
+    fstoggler.addEventListener("mouseover", this._expandCallback);
+    fstoggler.addEventListener("dragenter", this._expandCallback);
+    fstoggler.addEventListener("touchmove", this._expandCallback, {
+      passive: true,
+    });
 
-    if (enterFS) {
-      gNavToolbox.setAttribute("inFullscreen", true);
-      document.documentElement.setAttribute("inFullscreen", true);
-      let alwaysUsesNativeFullscreen =
+    document.documentElement.toggleAttribute("inFullscreen", enterFS);
+    document.documentElement.toggleAttribute(
+      "macOSNativeFullscreen",
+      enterFS &&
         AppConstants.platform == "macosx" &&
-        Services.prefs.getBoolPref("full-screen-api.macos-native-full-screen");
-      if (
-        (alwaysUsesNativeFullscreen || !document.fullscreenElement) &&
-        AppConstants.platform == "macosx"
-      ) {
-        document.documentElement.setAttribute("OSXLionFullscreen", true);
-      }
-    } else {
-      gNavToolbox.removeAttribute("inFullscreen");
-      document.documentElement.removeAttribute("inFullscreen");
-      document.documentElement.removeAttribute("OSXLionFullscreen");
-    }
+        (Services.prefs.getBoolPref(
+          "full-screen-api.macos-native-full-screen"
+        ) ||
+          !document.fullscreenElement)
+    );
 
     if (!document.fullscreenElement) {
-      this._updateToolbars(enterFS);
+      ToolbarIconColor.inferFromText("fullscreen", enterFS);
     }
 
     if (enterFS) {
       document.addEventListener("keypress", this._keyToggleCallback);
       document.addEventListener("popupshown", this._setPopupOpen);
       document.addEventListener("popuphidden", this._setPopupOpen);
-      gURLBar.controller.addQueryListener(this);
+      gURLBar.controller.addListener(this);
 
       // In DOM fullscreen mode, we hide toolbars with CSS
       if (!document.fullscreenElement) {
@@ -349,14 +397,52 @@ var FullScreen = {
       this._isPopupOpen = false;
       this.cleanup();
     }
+    this._toggleShortcutKeys();
   },
 
   exitDomFullScreen() {
+    // Don't use `this` here. It does not reliably refer to this object.
     if (document.fullscreen) {
       document.exitFullscreen();
     }
   },
 
+  _currentToolbarShift: 0,
+
+  /**
+   * Shifts the browser toolbar down when it is moused over on macOS in
+   * fullscreen.
+   * @param {number} shiftSize
+   *   A distance, in pixels, by which to shift the browser toolbar down.
+   */
+  shiftMacToolbarDown(shiftSize) {
+    if (typeof shiftSize !== "number") {
+      console.error("Tried to shift the toolbar by a non-numeric distance.");
+      return;
+    }
+
+    // shiftSize is sent from Cocoa widget code as a very precise double. We
+    // don't need that kind of precision in our CSS.
+    shiftSize = shiftSize.toFixed(2);
+    gNavToolbox.classList.toggle("fullscreen-with-menubar", shiftSize > 0);
+
+    let transform = shiftSize > 0 ? `translateY(${shiftSize}px)` : "";
+    gNavToolbox.style.transform = transform;
+    gURLBar.textbox.style.transform = gURLBar.textbox.hasAttribute("breakout")
+      ? transform
+      : "";
+    if (shiftSize > 0) {
+      // If the mouse tracking missed our fullScreenToggler, then the toolbox
+      // might not have been shown before the menubar is animated down. Make
+      // sure it is shown now.
+      if (!this.fullScreenToggler.hidden) {
+        this.showNavToolbox();
+      }
+    }
+
+    this._currentToolbarShift = shiftSize;
+  },
+
   handleEvent(event) {
     switch (event.type) {
       case "willenterfullscreen":
@@ -368,6 +454,9 @@ var FullScreen = {
       case "fullscreen":
         this.toggle();
         break;
+      case "MacFullscreenMenubarRevealUpdate":
+        this.shiftMacToolbarDown(event.detail);
+        break;
     }
   },
 
@@ -381,7 +470,6 @@ var FullScreen = {
     consoleMsg.initWithWindowID(
       message,
       gBrowser.currentURI.spec,
-      null,
       0,
       0,
       Ci.nsIScriptError.warningFlag,
@@ -412,7 +500,7 @@ var FullScreen = {
 
     // If we have a current pointerlock warning shown then hide it
     // before transition.
-    PointerlockFsWarning.close();
+    PointerlockFsWarning.close("pointerlock-warning");
 
     // If it is a remote browser, send a message to ask the content
     // to enter fullscreen state. We don't need to do so if it is an
@@ -427,21 +515,26 @@ var FullScreen = {
     // before the check is fine since we also check the activeness of
     // the requesting document in content-side handling code.
     if (this._isRemoteBrowser(aBrowser)) {
-      let [targetActor, inProcessBC] = this._getNextMsgRecipientActor(aActor);
+      // The cached message recipient in actor is used for fullscreen state
+      // cleanup, we should not use it while entering fullscreen.
+      let [targetActor, inProcessBC] = this._getNextMsgRecipientActor(
+        aActor,
+        false /* aUseCache */
+      );
       if (!targetActor) {
         // If there is no appropriate actor to send the message we have
         // no way to complete the transition and should abort by exiting
         // fullscreen.
-        this._abortEnterFullscreen();
+        this._abortEnterFullscreen(aActor);
         return;
       }
+      // Record that the actor is waiting for its child to enter
+      // fullscreen so that if it dies we can abort.
+      targetActor.waitingForChildEnterFullscreen = true;
       targetActor.sendAsyncMessage("DOMFullscreen:Entered", {
         remoteFrameBC: inProcessBC,
       });
 
-      // Record that the actor is waiting for its child to enter
-      // fullscreen so that if it dies we can abort.
-      targetActor.waitingForChildFullscreen = true;
       if (inProcessBC) {
         // We aren't messaging the request origin yet, skip this time.
         return;
@@ -459,7 +552,7 @@ var FullScreen = {
       // full-screen was made. Cancel full-screen.
       Services.focus.activeWindow != window
     ) {
-      this._abortEnterFullscreen();
+      this._abortEnterFullscreen(aActor);
       return;
     }
 
@@ -475,6 +568,8 @@ var FullScreen = {
     }
     document.documentElement.setAttribute("inDOMFullscreen", true);
 
+    XULBrowserWindow.onEnterDOMFullscreen();
+
     if (gFindBarInitialized) {
       gFindBar.close(true);
     }
@@ -503,7 +598,26 @@ var FullScreen = {
       document.removeEventListener("keypress", this._keyToggleCallback);
       document.removeEventListener("popupshown", this._setPopupOpen);
       document.removeEventListener("popuphidden", this._setPopupOpen);
-      gURLBar.controller.removeQueryListener(this);
+      gURLBar.controller.removeListener(this);
+    }
+  },
+
+  _toggleShortcutKeys() {
+    const kEnterKeyIds = [
+      "key_enterFullScreen",
+      "key_enterFullScreen_old",
+      "key_enterFullScreen_compat",
+    ];
+    const kExitKeyIds = [
+      "key_exitFullScreen",
+      "key_exitFullScreen_old",
+      "key_exitFullScreen_compat",
+    ];
+    for (let id of window.fullScreen ? kEnterKeyIds : kExitKeyIds) {
+      document.getElementById(id)?.setAttribute("disabled", "true");
+    }
+    for (let id of window.fullScreen ? kExitKeyIds : kEnterKeyIds) {
+      document.getElementById(id)?.removeAttribute("disabled");
     }
   },
 
@@ -518,13 +632,24 @@ var FullScreen = {
    * the cleanup.
    */
   cleanupDomFullscreen(aActor) {
-    let [target, inProcessBC] = this._getNextMsgRecipientActor(aActor);
+    let needToWaitForChildExit = false;
+    // Use the message recipient cached in the actor if possible, especially for
+    // the case that actor is destroyed, which we are unable to find it by
+    // walking up the browsing context tree.
+    let [target, inProcessBC] = this._getNextMsgRecipientActor(
+      aActor,
+      true /* aUseCache */
+    );
     if (target) {
+      needToWaitForChildExit = true;
+      // Record that the actor is waiting for its child to exit fullscreen so
+      // that if it dies we can continue cleanup.
+      target.waitingForChildExitFullscreen = true;
       target.sendAsyncMessage("DOMFullscreen:CleanUp", {
         remoteFrameBC: inProcessBC,
       });
       if (inProcessBC) {
-        return;
+        return needToWaitForChildExit;
       }
     }
 
@@ -534,23 +659,29 @@ var FullScreen = {
       true
     );
 
-    PointerlockFsWarning.close();
+    PointerlockFsWarning.close("fullscreen-warning");
     gBrowser.tabContainer.removeEventListener(
       "TabSelect",
       this.exitDomFullScreen
     );
 
     document.documentElement.removeAttribute("inDOMFullscreen");
+
+    return needToWaitForChildExit;
   },
 
-  _abortEnterFullscreen() {
+  _abortEnterFullscreen(aActor) {
     // This function is called synchronously in fullscreen change, so
     // we have to avoid calling exitFullscreen synchronously here.
-    setTimeout(() => document.exitFullscreen(), 0);
-    if (TelemetryStopwatch.running("FULLSCREEN_CHANGE_MS")) {
+    //
+    // This could reject if we're not currently in fullscreen
+    // so just ignore rejection.
+    setTimeout(() => document.exitFullscreen().catch(() => {}), 0);
+    if (aActor.timerId) {
       // Cancel the stopwatch for any fullscreen change to avoid
       // errors if it is started again.
-      TelemetryStopwatch.cancel("FULLSCREEN_CHANGE_MS");
+      Glean.fullscreen.change.cancel(aActor.timerId);
+      aActor.timerId = null;
     }
   },
 
@@ -562,15 +693,35 @@ var FullScreen = {
    *
    * @param {JSWindowActorParent} aActor
    *        The actor that called this function.
+   * @param {bool} aUseCache
+   *        Use the recipient cached in the aActor if available.
    *
    * @return {[JSWindowActorParent, BrowsingContext]}
    *         The parent actor which should be sent the next msg and the
    *         in process browsing context which is its child. Will be
    *         [null, null] if there is no OOP parent actor and request origin
    *         is unset. [null, null] is also returned if the intended actor or
-   *         the calling actor has been destroyed.
+   *         the calling actor has been destroyed or its associated
+   *         WindowContext is in BFCache.
    */
-  _getNextMsgRecipientActor(aActor) {
+  _getNextMsgRecipientActor(aActor, aUseCache) {
+    // Walk up the cached nextMsgRecipient to find the next available actor if
+    // any.
+    if (aUseCache && aActor.nextMsgRecipient) {
+      let nextMsgRecipient = aActor.nextMsgRecipient;
+      while (nextMsgRecipient) {
+        let [actor] = nextMsgRecipient;
+        if (
+          !actor.hasBeenDestroyed() &&
+          actor.windowContext &&
+          !actor.windowContext.isInBFCache
+        ) {
+          return nextMsgRecipient;
+        }
+        nextMsgRecipient = actor.nextMsgRecipient;
+      }
+    }
+
     if (aActor.hasBeenDestroyed()) {
       return [null, null];
     }
@@ -601,11 +752,16 @@ var FullScreen = {
     if (parentBC && parentBC.currentWindowGlobal) {
       target = parentBC.currentWindowGlobal.getActor("DOMFullscreen");
       inProcessBC = childBC;
+      aActor.nextMsgRecipient = [target, inProcessBC];
     } else {
       target = aActor.requestOrigin;
     }
 
-    if (!target || target.hasBeenDestroyed()) {
+    if (
+      !target ||
+      target.hasBeenDestroyed() ||
+      target.windowContext?.isInBFCache
+    ) {
       return [null, null];
     }
     return [target, inProcessBC];
@@ -649,7 +805,7 @@ var FullScreen = {
     // e.g. we wouldn't want the autoscroll icon firing this event, so when the user
     // toggles chrome when moving mouse to the top, it doesn't go away again.
     let target = aEvent.originalTarget;
-    if (target.localName == "tooltip") {
+    if (target.localName == "tooltip" || target.id == "tab-preview-panel") {
       return;
     }
     if (
@@ -683,7 +839,7 @@ var FullScreen = {
   },
 
   // Autohide helpers for the context menu item
-  getAutohide(aItem) {
+  updateAutohideMenuitem(aItem) {
     aItem.setAttribute(
       "checked",
       Services.prefs.getBoolPref("browser.fullscreen.autohide")
@@ -702,7 +858,7 @@ var FullScreen = {
     if (BrowserHandler.kiosk) {
       return;
     }
-    this._fullScrToggler.hidden = true;
+    this.fullScreenToggler.hidden = true;
     gNavToolbox.removeAttribute("fullscreenShouldAnimate");
     gNavToolbox.style.marginTop = "";
 
@@ -711,7 +867,7 @@ var FullScreen = {
     }
 
     // Track whether mouse is near the toolbox
-    if (trackMouse && AppConstants.platform != "macosx") {
+    if (trackMouse) {
       let rect = gBrowser.tabpanels.getBoundingClientRect();
       this._mouseTargetRect = {
         top: rect.top + 50,
@@ -737,10 +893,6 @@ var FullScreen = {
     if (this._isPopupOpen) {
       return;
     }
-    // On macOS we don't want to hide toolbars.
-    if (AppConstants.platform == "macosx") {
-      return;
-    }
 
     // a textbox in chrome is focused (location bar anyone?): don't collapse chrome
     // unless we are kiosk mode
@@ -777,7 +929,7 @@ var FullScreen = {
     }
 
     if (!BrowserHandler.kiosk) {
-      this._fullScrToggler.hidden = false;
+      this.fullScreenToggler.hidden = false;
     }
 
     if (
@@ -795,65 +947,27 @@ var FullScreen = {
 
     MousePosTracker.removeListener(this);
   },
-
-  _updateToolbars(aEnterFS) {
-    for (let el of document.querySelectorAll(
-      "toolbar[fullscreentoolbar=true]"
-    )) {
-      if (aEnterFS) {
-        // Give the main nav bar and the tab bar the fullscreen context menu,
-        // otherwise remove context menu to prevent breakage
-        el.setAttribute("saved-context", el.getAttribute("context"));
-        if (el.id == "nav-bar" || el.id == "TabsToolbar") {
-          el.setAttribute("context", "autohide-context");
-        } else {
-          el.removeAttribute("context");
-        }
-
-        // Set the inFullscreen attribute to allow specific styling
-        // in fullscreen mode
-        el.setAttribute("inFullscreen", true);
-      } else {
-        if (el.hasAttribute("saved-context")) {
-          el.setAttribute("context", el.getAttribute("saved-context"));
-          el.removeAttribute("saved-context");
-        }
-        el.removeAttribute("inFullscreen");
-      }
-    }
-
-    ToolbarIconColor.inferFromText("fullscreen", aEnterFS);
-
-    // For macOS, we use native full screen, all full screen controls
-    // are hidden, don't bother to touch them. If we don't stop here,
-    // the following code could cause the native full screen button be
-    // shown unexpectedly. See bug 1165570.
-    if (AppConstants.platform == "macosx") {
-      return;
-    }
-
-    var fullscreenctls = document.getElementById("window-controls");
-    var navbar = document.getElementById("nav-bar");
-    var ctlsOnTabbar = window.toolbar.visible;
-    if (fullscreenctls.parentNode == navbar && ctlsOnTabbar) {
-      fullscreenctls.removeAttribute("flex");
-      document.getElementById("TabsToolbar").appendChild(fullscreenctls);
-    } else if (fullscreenctls.parentNode.id == "TabsToolbar" && !ctlsOnTabbar) {
-      fullscreenctls.setAttribute("flex", "1");
-      navbar.appendChild(fullscreenctls);
-    }
-    fullscreenctls.hidden = !aEnterFS;
-  },
 };
 
-XPCOMUtils.defineLazyGetter(FullScreen, "_permissionNotificationIDs", () => {
-  let { PermissionUI } = ChromeUtils.import(
-    "resource:///modules/PermissionUI.jsm",
-    {}
+ChromeUtils.defineLazyGetter(FullScreen, "_permissionNotificationIDs", () => {
+  let { PermissionUI } = ChromeUtils.importESModule(
+    "resource:///modules/PermissionUI.sys.mjs"
   );
   return (
     Object.values(PermissionUI)
-      .filter(value => value.prototype && value.prototype.notificationID)
+      .filter(value => {
+        let returnValue;
+        try {
+          returnValue = value.prototype.notificationID;
+        } catch (err) {
+          if (err.message === "Not implemented.") {
+            returnValue = false;
+          } else {
+            throw err;
+          }
+        }
+        return returnValue;
+      })
       .map(value => value.prototype.notificationID)
       // Additionally include webRTC permission prompt which does not use PermissionUI
       .concat(["webRTC-shareDevices"])
