@@ -13,39 +13,119 @@ var isDevtools = SimpleTest.harnessParameters.subsuite == "devtools";
 // This list should contain only path prefixes. It is meant to stop the test
 // from reporting things that *are* referenced, but for which the test can't
 // find any reference because the URIs are constructed programatically.
-// If you need to whitelist specific files, please use the 'whitelist' object.
+// If you need to allowlist specific files, please use the 'allowlist' object.
 var gExceptionPaths = [
-  "chrome://browser/content/defaultthemes/",
   "resource://app/defaults/settings/blocklists/",
   "resource://app/defaults/settings/security-state/",
   "resource://app/defaults/settings/main/",
-  "resource://app/defaults/settings/pinning/",
   "resource://app/defaults/preferences/",
   "resource://gre/modules/commonjs/",
   "resource://gre/defaults/pref/",
 
-  // These resources are referenced using relative paths from html files.
-  "resource://payments/",
+  // These chrome resources are referenced using relative paths from JS files.
+  "chrome://global/content/certviewer/components/",
 
   // https://github.com/mozilla/activity-stream/issues/3053
   "chrome://activity-stream/content/data/content/tippytop/images/",
   "chrome://activity-stream/content/data/content/tippytop/favicons/",
   // These resources are referenced by messages delivered through Remote Settings
-  "chrome://activity-stream/content/data/content/assets/remote/",
+  "chrome://activity-stream/content/data/content/assets/mobile-download-qr-new-user-cn.svg",
+  "chrome://activity-stream/content/data/content/assets/mobile-download-qr-existing-user-cn.svg",
+  "chrome://activity-stream/content/data/content/assets/mr-amo-collection.svg",
+  "chrome://activity-stream/content/data/content/assets/person-typing.svg",
+  "chrome://activity-stream/content/data/content/assets/tabs-side-zap-transparent.svg",
+  "chrome://activity-stream/content/data/content/assets/tabs-top-zap-transparent.svg",
+  "chrome://activity-stream/content/data/content/assets/nuo-taborientation.svg",
+  "chrome://activity-stream/content/data/content/assets/euo-tab-orientation.svg",
+  "chrome://activity-stream/content/data/content/assets/euo-chatbot.svg",
+  "chrome://browser/content/assets/moz-vpn.svg",
+  "chrome://browser/content/assets/vpn-logo.svg",
+  "chrome://browser/content/assets/focus-promo.png",
+  "chrome://browser/content/assets/klar-qr-code.svg",
+  "chrome://browser/content/asrouter/assets/fox-with-box-on-cloud.svg",
+  "chrome://browser/content/asrouter/assets/fox-with-devices.svg",
+  "chrome://browser/content/asrouter/assets/fox-with-locked-box.svg",
+  "chrome://browser/content/asrouter/assets/fox-with-mobile.svg",
 
   // toolkit/components/pdfjs/content/build/pdf.js
   "resource://pdf.js/web/images/",
+  // This file is only loaded in using a dynamic import in pdf.js in case wasm
+  // is not available.
+  "resource://pdf.js/web/wasm/openjpeg_nowasm_fallback.js",
 
-  // Exclude all the metadata paths under the country metadata folder because these
-  // paths will be concatenated in FormAutofillUtils.jsm based on different country/region.
-  "resource://formautofill/addressmetadata/",
+  // Exclude the form autofill path that has been moved out of the extensions to
+  // toolkit, see bug 1691821.
+  "resource://gre-resources/autofill/",
+  // Localization file added programatically in FormAutofillUtils.sys.mjs
+  "resource://gre/localization/en-US/toolkit/formautofill",
 
   // Exclude all search-extensions because they aren't referenced by filename
   "resource://search-extensions/",
 
   // Exclude all services-automation because they are used through webdriver
   "resource://gre/modules/services-automation/",
-  "resource://services-automation/ServicesAutomation.jsm",
+
+  // Paths from this folder are constructed in NetErrorParent.sys.mjs based on
+  // the type of cert or net error the user is encountering.
+  "chrome://global/content/neterror/supportpages/",
+
+  // Points to theme preview images, which are defined in browser/ but only used
+  // in toolkit/mozapps/extensions/content/aboutaddons.js.
+  "resource://usercontext-content/builtin-themes/",
+
+  // Page data schemas are referenced programmatically.
+  "chrome://browser/content/pagedata/schemas/",
+
+  // Nimbus schemas are referenced programmatically.
+  "resource://nimbus/schemas/",
+
+  // Normandy schemas are referenced programmatically.
+  "resource://normandy/schemas/",
+
+  // ASRouter schemas are referenced programmatically.
+  "chrome://browser/content/asrouter/schemas/",
+
+  // Localization file added programatically in FeatureCallout.sys.mjs
+  "resource://app/localization/en-US/browser/featureCallout.ftl",
+
+  // Localization file added programatically in ContentAnalysis.sys.mjs
+  "resource://gre/localization/en-US/toolkit/contentanalysis/",
+
+  // CSS files are referenced inside JS in an html template
+  "chrome://browser/content/aboutlogins/components/",
+
+  // Strip on Share parameter lists
+  "chrome://global/content/antitracking/",
+
+  // CSS file is referenced inside JS in login-form.mjs
+  "chrome://global/content/megalist/LoginFormComponent/",
+
+  // The ONNX runtime picks files to run programmaticaly
+  "chrome://global/content/ml/",
+
+  // The profile avatars are directly referenced.
+  "chrome://browser/content/profiles/assets/",
+
+  // The picture-in-picture add-on.
+  "resource://builtin-addons/pictureinpicture/",
+
+  // The formautofill add-on.
+  "resource://builtin-addons/formautofill/",
+
+  // The webcompat add-on.
+  "resource://builtin-addons/webcompat/",
+
+  // The newtab add-on
+  "resource://builtin-addons/newtab/",
+  "resource://newtab/",
+  "chrome://newtab/",
+
+  // UniFFI test files.
+  "moz-src:///toolkit/components/uniffi-bindgen-gecko-js/tests/generated/",
+
+  // Used for Market suggestions on the urlbar. This is specified from Remote
+  // Settings.
+  "chrome://browser/skin/illustrations/market-opt-in.svg",
 ];
 
 // These are not part of the omni.ja file, so we find them only when running
@@ -55,20 +135,26 @@ if (AppConstants.platform == "macosx") {
   gExceptionPaths.push("resource://gre/res/touchbar/");
 }
 
-// Each whitelist entry should have a comment indicating which file is
-// referencing the whitelisted file in a way that the test can't detect, or a
+if (AppConstants.MOZ_BACKGROUNDTASKS) {
+  // These preferences are active only when we're in background task mode.
+  gExceptionPaths.push("resource://gre/defaults/backgroundtasks/");
+  gExceptionPaths.push("resource://app/defaults/backgroundtasks/");
+  // `BackgroundTask_*.sys.mjs` are loaded at runtime by `app --backgroundtask id ...`.
+  gExceptionPaths.push("resource://gre/modules/backgroundtasks/");
+  gExceptionPaths.push("resource://app/modules/backgroundtasks/");
+}
+
+// Each allowlist entry should have a comment indicating which file is
+// referencing the listed file in a way that the test can't detect, or a
 // bug number to remove or use the file if it is indeed currently unreferenced.
-var whitelist = [
-  // pocket/content/panels/tmpl/loggedoutvariants/variant_a.handlebars
-  { file: "chrome://pocket/content/panels/img/glyph.svg" },
-
-  // toolkit/components/pdfjs/content/PdfStreamConverter.jsm
-  { file: "chrome://pdf.js/locale/chrome.properties" },
-  { file: "chrome://pdf.js/locale/viewer.properties" },
-
+var allowlist = [
   // security/manager/pki/resources/content/device_manager.js
   { file: "chrome://pippki/content/load_device.xhtml" },
 
+  // Intentionally unreferenced, see bug 1941134
+  { file: "resource://gre/res/designmode.css" },
+  { file: "resource://gre/res/EditorOverride.css" },
+
   // The l10n build system can't package string files only for some platforms.
   // See bug 1339424 for why this is hard to fix.
   {
@@ -77,11 +163,14 @@ var whitelist = [
   },
   {
     file: "resource://gre/localization/en-US/toolkit/printing/printDialogs.ftl",
-    platforms: ["macosx"],
+    platforms: ["linux", "macosx"],
   },
 
-  // toolkit/content/aboutRights-unbranded.xhtml doesn't use aboutRights.css
-  { file: "chrome://global/skin/aboutRights.css", skipUnofficial: true },
+  // This file is referenced by the build system to generate the
+  // Firefox .desktop entry. See bug 1824327 (and perhaps bug 1526672)
+  {
+    file: "resource://app/localization/en-US/browser/linuxDesktopEntry.ftl",
+  },
 
   // devtools/client/inspector/bin/dev-server.js
   {
@@ -89,100 +178,91 @@ var whitelist = [
     isFromDevTools: true,
   },
 
+  // used by devtools/client/memory/index.xhtml
+  { file: "chrome://global/content/third_party/d3/d3.js" },
+
   // SpiderMonkey parser API, currently unused in browser/ and toolkit/
-  { file: "resource://gre/modules/reflect.jsm" },
+  { file: "resource://gre/modules/reflect.sys.mjs" },
 
   // extensions/pref/autoconfig/src/nsReadConfig.cpp
   { file: "resource://gre/defaults/autoconfig/prefcalls.js" },
 
-  // modules/libpref/Preferences.cpp
+  // browser/components/preferences/moreFromMozilla.js
+  // These files URLs are constructed programatically at run time.
+  {
+    file: "chrome://browser/content/preferences/more-from-mozilla-qr-code-simple.svg",
+  },
+  {
+    file: "chrome://browser/content/preferences/more-from-mozilla-qr-code-simple-cn.svg",
+  },
+
   { file: "resource://gre/greprefs.js" },
 
   // layout/mathml/nsMathMLChar.cpp
   { file: "resource://gre/res/fonts/mathfontSTIXGeneral.properties" },
   { file: "resource://gre/res/fonts/mathfontUnicode.properties" },
 
+  // toolkit/mozapps/extensions/AddonContentPolicy.cpp
+  { file: "resource://gre/localization/en-US/toolkit/global/cspErrors.ftl" },
+
+  // toolkit/components/antitracking/bouncetrackingprotection/BounceTrackingProtection.cpp
+  { file: "resource://gre/localization/en-US/toolkit/global/antiTracking.ftl" },
+
   // The l10n build system can't package string files only for some platforms.
   {
-    file:
-      "resource://gre/chrome/en-US/locale/en-US/global-platform/mac/accessible.properties",
+    file: "resource://gre/chrome/en-US/locale/en-US/global-platform/mac/accessible.properties",
     platforms: ["linux", "win"],
   },
   {
-    file:
-      "resource://gre/chrome/en-US/locale/en-US/global-platform/mac/intl.properties",
+    file: "resource://gre/chrome/en-US/locale/en-US/global-platform/mac/intl.properties",
     platforms: ["linux", "win"],
   },
   {
-    file:
-      "resource://gre/chrome/en-US/locale/en-US/global-platform/mac/platformKeys.properties",
+    file: "resource://gre/chrome/en-US/locale/en-US/global-platform/mac/platformKeys.properties",
     platforms: ["linux", "win"],
   },
   {
-    file:
-      "resource://gre/chrome/en-US/locale/en-US/global-platform/unix/accessible.properties",
+    file: "resource://gre/chrome/en-US/locale/en-US/global-platform/unix/accessible.properties",
     platforms: ["macosx", "win"],
   },
   {
-    file:
-      "resource://gre/chrome/en-US/locale/en-US/global-platform/unix/intl.properties",
+    file: "resource://gre/chrome/en-US/locale/en-US/global-platform/unix/intl.properties",
     platforms: ["macosx", "win"],
   },
   {
-    file:
-      "resource://gre/chrome/en-US/locale/en-US/global-platform/unix/platformKeys.properties",
+    file: "resource://gre/chrome/en-US/locale/en-US/global-platform/unix/platformKeys.properties",
     platforms: ["macosx", "win"],
   },
   {
-    file:
-      "resource://gre/chrome/en-US/locale/en-US/global-platform/win/accessible.properties",
+    file: "resource://gre/chrome/en-US/locale/en-US/global-platform/win/accessible.properties",
     platforms: ["linux", "macosx"],
   },
   {
-    file:
-      "resource://gre/chrome/en-US/locale/en-US/global-platform/win/intl.properties",
+    file: "resource://gre/chrome/en-US/locale/en-US/global-platform/win/intl.properties",
     platforms: ["linux", "macosx"],
   },
   {
-    file:
-      "resource://gre/chrome/en-US/locale/en-US/global-platform/win/platformKeys.properties",
+    file: "resource://gre/chrome/en-US/locale/en-US/global-platform/win/platformKeys.properties",
     platforms: ["linux", "macosx"],
   },
 
   // Files from upstream library
-  { file: "resource://pdf.js/build/pdf.sandbox.external.js" },
-  { file: "resource://pdf.js/build/pdf.scripting.js" },
-  { file: "resource://pdf.js/web/debugger.js" },
+  { file: "resource://pdf.js/web/debugger.mjs" },
+  { file: "resource://pdf.js/web/debugger.css" },
 
-  // resource://app/modules/translation/TranslationContentHandler.jsm
-  { file: "resource://app/modules/translation/BingTranslator.jsm" },
-  { file: "resource://app/modules/translation/GoogleTranslator.jsm" },
-  { file: "resource://app/modules/translation/YandexTranslator.jsm" },
+  // File from the ipp-activator add-on
+  { file: "resource://builtin-addons/ipp-activator/breakages/tab.json" },
 
-  // Starting from here, files in the whitelist are bugs that need fixing.
+  // Starting from here, files in the allowlist are bugs that need fixing.
   // Bug 1339424 (wontfix?)
   {
     file: "chrome://browser/locale/taskbar.properties",
     platforms: ["linux", "macosx"],
   },
-  // Bug 1619090 to clean up platform-specific crypto
-  {
-    file: "resource://gre/modules/OSCrypto.jsm",
-    platforms: ["linux", "macosx"],
-  },
-  // Bug 1344267
-  { file: "chrome://marionette/content/test.xhtml" },
-  { file: "chrome://marionette/content/test_dialog.properties" },
-  { file: "chrome://marionette/content/test_dialog.xhtml" },
-  { file: "chrome://marionette/content/test_menupopup.xhtml" },
   // Bug 1348559
   { file: "chrome://pippki/content/resetpassword.xhtml" },
   // Bug 1337345
-  { file: "resource://gre/modules/Manifest.jsm" },
-  // Bug 1356045
-  { file: "chrome://global/content/test-ipc.xhtml" },
-  // Bug 1378173 (warning: still used by devtools)
-  { file: "resource://gre/modules/Promise.jsm" },
+  { file: "resource://gre/modules/Manifest.sys.mjs" },
   // Bug 1494170
   // (The references to these files are dynamically generated, so the test can't
   // find the references)
@@ -205,40 +285,105 @@ var whitelist = [
     file: "resource://app/localization/en-US/browser/touchbar/touchbar.ftl",
     platforms: ["linux", "win"],
   },
-  // Referenced by the webcompat system addon for localization
-  { file: "resource://gre/localization/en-US/toolkit/about/aboutCompat.ftl" },
-
-  // Bug 1559554
-  { file: "chrome://browser/content/aboutlogins/aboutLoginsUtils.js" },
-
-  // Referenced from the screenshots webextension
-  { file: "resource://app/localization/en-US/browser/screenshots.ftl" },
-
-  // services/fxaccounts/RustFxAccount.js
-  { file: "resource://gre/modules/RustFxAccount.js" },
 
   // dom/media/mediacontrol/MediaControlService.cpp
   { file: "resource://gre/localization/en-US/dom/media.ftl" },
+
+  // dom/xml/nsXMLPrettyPrinter.cpp
+  { file: "resource://gre/localization/en-US/dom/XMLPrettyPrint.ftl" },
+
+  // tookit/mozapps/update/BackgroundUpdate.sys.mjs
+  {
+    file: "resource://gre/localization/en-US/toolkit/updates/backgroundupdate.ftl",
+  },
+
+  // Bug 1713242 - referenced by aboutThirdParty.html which is only for Windows
+  {
+    file: "resource://gre/localization/en-US/toolkit/about/aboutThirdParty.ftl",
+    platforms: ["linux", "macosx"],
+  },
+  // Bug 1854618 - referenced by aboutWebauthn.html which is only for Linux and Mac
+  {
+    file: "resource://gre/localization/en-US/toolkit/about/aboutWebauthn.ftl",
+    platforms: ["win", "android"],
+  },
+  // Bug 1973834 - referenced by aboutWindowsMessages.html which is only for Windows
+  {
+    file: "resource://gre/localization/en-US/toolkit/about/aboutWindowsMessages.ftl",
+    platforms: ["linux", "macosx"],
+  },
+  // Bug 1721741:
+  // (The references to these files are dynamically generated, so the test can't
+  // find the references)
+  { file: "chrome://browser/content/screenshots/copied-notification.svg" },
+
+  // toolkit/xre/MacRunFromDmgUtils.mm
+  { file: "resource://gre/localization/en-US/toolkit/global/run-from-dmg.ftl" },
+
+  // Referenced programmatically
+  { file: "chrome://browser/content/backup/BackupManifest.1.schema.json" },
+  { file: "chrome://browser/content/backup/ArchiveJSONBlock.1.schema.json" },
+
+  // Bug 1733498 - Migrate necko errors l10n strings from .properties to Fluent
+  {
+    file: "resource://gre/localization/en-US/netwerk/necko.ftl",
+  },
+
+  // dom/xslt/xslt/txMozillaXSLTProcessor.cpp
+  { file: "resource://gre/localization/en-US/dom/xslt.ftl" },
+
+  // A QA and dev debug tool.
+  { file: "chrome://browser/content/places/interactionsViewer.html" },
+
+  // A uniffi-generated file which isn't used directly by 1969013 but will be by a followup.
+  {
+    file: "moz-src:///toolkit/components/uniffi-bindgen-gecko-js/components/generated/RustTracing.sys.mjs",
+  },
 ];
 
-if (AppConstants.NIGHTLY_BUILD && AppConstants.platform != "win") {
-  // This path is refereneced in nsFxrCommandLineHandler.cpp, which is only
-  // compiled in Windows. Whitelisted this path so that non-Windows builds
-  // can access the FxR UI via --chrome rather than --fxr (which includes VR-
-  // specific functionality)
-  whitelist.push({ file: "chrome://fxr/content/fxrui.html" });
+if (AppConstants.NIGHTLY_BUILD) {
+  allowlist.push(
+    // A debug tool that is only available in Nightly builds, and is accessed
+    // directly by developers via the chrome URI (bug 1888491)
+    { file: "chrome://browser/content/backup/debug.html" }
+  );
+}
+
+if (AppConstants.platform != "win") {
+  // toolkit/mozapps/defaultagent/Notification.cpp
+  // toolkit/mozapps/defaultagent/ScheduledTask.cpp
+  // toolkit/mozapps/defaultagent/BackgroundTask_defaultagent.sys.mjs
+  // Bug 1854425 - referenced by default browser agent which is not detected
+  allowlist.push({
+    file: "resource://app/localization/en-US/browser/backgroundtasks/defaultagent.ftl",
+  });
+
+  if (AppConstants.NIGHTLY_BUILD) {
+    // This path is refereneced in nsFxrCommandLineHandler.cpp, which is only
+    // compiled in Windows. This path is allowed so that non-Windows builds
+    // can access the FxR UI via --chrome rather than --fxr (which includes VR-
+    // specific functionality)
+    allowlist.push({ file: "chrome://fxr/content/fxrui.html" });
+  }
 }
 
 if (AppConstants.platform == "android") {
   // The l10n build system can't package string files only for some platforms.
   // Referenced by aboutGlean.html
-  whitelist.push({
+  allowlist.push({
     file: "resource://gre/localization/en-US/toolkit/about/aboutGlean.ftl",
   });
 }
 
-whitelist = new Set(
-  whitelist
+if (AppConstants.MOZ_UPDATE_AGENT && !AppConstants.MOZ_BACKGROUNDTASKS) {
+  // Task scheduling is only used for background updates right now.
+  allowlist.push({
+    file: "resource://gre/modules/TaskScheduler.sys.mjs",
+  });
+}
+
+allowlist = new Set(
+  allowlist
     .filter(
       item =>
         "isFromDevTools" in item == isDevtools &&
@@ -248,52 +393,45 @@ whitelist = new Set(
     .map(item => item.file)
 );
 
-const ignorableWhitelist = new Set([
+const ignorableAllowlist = new Set([
   // The following files are outside of the omni.ja file, so we only catch them
   // when testing on a non-packaged build.
 
-  // toolkit/mozapps/extensions/nsBlocklistService.js
-  "resource://app/blocklist.xml",
-
   // dom/media/gmp/GMPParent.cpp
   "resource://gre/gmp-clearkey/0.1/manifest.json",
-
-  // Bug 1351669 - obsolete test file
-  "resource://gre/res/test.properties",
 ]);
-for (let entry of ignorableWhitelist) {
-  whitelist.add(entry);
+for (let entry of ignorableAllowlist) {
+  allowlist.add(entry);
 }
 
 if (!isDevtools) {
-  // services/sync/modules/main.js
-  whitelist.add("resource://services-sync/service.js");
-  // services/sync/modules/service.js
+  // services/sync/modules/service.sys.mjs
   for (let module of [
-    "addons.js",
-    "bookmarks.js",
-    "forms.js",
-    "history.js",
-    "passwords.js",
-    "prefs.js",
-    "tabs.js",
-    "extension-storage.js",
+    "addons.sys.mjs",
+    "bookmarks.sys.mjs",
+    "forms.sys.mjs",
+    "history.sys.mjs",
+    "passwords.sys.mjs",
+    "prefs.sys.mjs",
+    "tabs.sys.mjs",
+    "extension-storage.sys.mjs",
   ]) {
-    whitelist.add("resource://services-sync/engines/" + module);
+    allowlist.add("resource://services-sync/engines/" + module);
   }
   // resource://devtools/shared/worker/loader.js,
-  // resource://devtools/shared/builtin-modules.js
-  if (!AppConstants.ENABLE_REMOTE_AGENT) {
-    whitelist.add("resource://gre/modules/jsdebugger.jsm");
+  // resource://devtools/shared/loader/builtin-modules.js
+  if (!AppConstants.ENABLE_WEBDRIVER) {
+    allowlist.add("resource://gre/modules/jsdebugger.sys.mjs");
   }
 }
 
 if (AppConstants.MOZ_CODE_COVERAGE) {
-  whitelist.add("chrome://marionette/content/PerTestCoverageUtils.jsm");
+  allowlist.add(
+    "chrome://remote/content/marionette/PerTestCoverageUtils.sys.mjs"
+  );
 }
 
 const gInterestingCategories = new Set([
-  "agent-style-sheets",
   "addon-provider-module",
   "webextension-modules",
   "webextension-scripts",
@@ -308,14 +446,13 @@ var gChromeReg = Cc["@mozilla.org/chrome/chrome-registry;1"].getService(
 );
 var gChromeMap = new Map();
 var gOverrideMap = new Map();
-var gComponentsSet = new Set();
 
 // In this map when the value is a Set of URLs, the file is referenced if any
 // of the files in the Set is referenced.
 // When the value is null, the file is referenced unconditionally.
-// When the value is a string, "whitelist-direct" means that we have not found
-// any reference in the code, but have a matching whitelist entry for this file.
-// "whitelist" means that the file is indirectly whitelisted, ie. a whitelisted
+// When the value is a string, "allowlist-direct" means that we have not found
+// any reference in the code, but have a matching allowlist entry for this file.
+// "allowlist" means that the file is indirectly allowlisted, ie. a allowlisted
 // file causes this file to be referenced.
 var gReferencesFromCode = new Map();
 
@@ -331,7 +468,7 @@ trackResourcePrefix("gre");
 trackResourcePrefix("app");
 
 function getBaseUriForChromeUri(chromeUri) {
-  let chromeFile = chromeUri + "gobbledygooknonexistentfile.reallynothere";
+  let chromeFile = chromeUri + "nonexistentfile.reallynothere";
   let uri = Services.io.newURI(chromeFile);
   let fileUri = gChromeReg.convertChromeURL(uri);
   return fileUri.resolve(".");
@@ -344,7 +481,7 @@ function trackChromeUri(uri) {
 // formautofill registers resource://formautofill/ and
 // chrome://formautofill/content/ dynamically at runtime.
 // Bug 1480276 is about addressing this without this hard-coding.
-trackResourcePrefix("formautofill");
+trackResourcePrefix("autofill");
 trackChromeUri("chrome://formautofill/content/");
 
 function parseManifest(manifestUri) {
@@ -353,17 +490,7 @@ function parseManifest(manifestUri) {
       let [type, ...argv] = line.split(/\s+/);
       if (type == "content" || type == "skin" || type == "locale") {
         let chromeUri = `chrome://${argv[0]}/${type}/`;
-        // The webcompat reporter's locale directory may not exist if
-        // the addon is preffed-off, and since it's a hack until we
-        // get bz1425104 landed, we'll just skip it for now.
-        // Same issue with fxmonitor, which is also pref'd off.
-        if (chromeUri === "chrome://report-site-issue/locale/") {
-          gChromeMap.set("chrome://report-site-issue/locale/", true);
-        } else if (chromeUri === "chrome://fxmonitor/locale/") {
-          gChromeMap.set("chrome://fxmonitor/locale/", true);
-        } else {
-          trackChromeUri(chromeUri);
-        }
+        trackChromeUri(chromeUri);
       } else if (type == "override" || type == "overlay") {
         // Overlays aren't really overrides, but behave the same in
         // that the overlay is only referenced if the original xul
@@ -375,12 +502,20 @@ function parseManifest(manifestUri) {
             Services.io.newURI(argv[0]).specIgnoringRef
           );
         }
-      } else if (type == "category" && gInterestingCategories.has(argv[0])) {
-        gReferencesFromCode.set(argv[2], null);
+      } else if (type == "category") {
+        if (gInterestingCategories.has(argv[0])) {
+          gReferencesFromCode.set(argv[2], null);
+        } else if (
+          argv[1].startsWith("resource://") ||
+          argv[1].startsWith("moz-src://")
+        ) {
+          // Assume that any resource paths immediately after the category name
+          // are for use with BrowserUtils.callModulesFromCategory (rather than
+          // having to hardcode a list of categories in this test).
+          gReferencesFromCode.set(argv[1], null);
+        }
       } else if (type == "resource") {
         trackResourcePrefix(argv[0]);
-      } else if (type == "component") {
-        gComponentsSet.add(argv[1]);
       }
     }
   });
@@ -406,6 +541,12 @@ async function parseJsonManifest(uri) {
     return uri;
   }
 
+  if (data.background?.scripts) {
+    for (let bgscript of data.background.scripts) {
+      gReferencesFromCode.set(uri.resolve(bgscript), null);
+    }
+  }
+
   if (data.icons) {
     for (let icon of Object.values(data.icons)) {
       gReferencesFromCode.set(uri.resolve(icon), null);
@@ -418,6 +559,10 @@ async function parseJsonManifest(uri) {
         let script = uri.resolve(api.parent.script);
         gReferencesFromCode.set(script, null);
       }
+
+      if (api.schema) {
+        gReferencesFromCode.set(uri.resolve(api.schema), null);
+      }
     }
   }
 
@@ -511,6 +656,10 @@ function parseCodeFile(fileUri) {
         /["'`]chrome:\/\/[a-zA-Z0-9-]+\/(content|skin|locale)\/[^"'` ]*["'`]/g
       );
 
+      if (!urls) {
+        urls = line.match(/["']moz-src:\/\/[^"']+["']/g);
+      }
+
       if (!urls) {
         urls = line.match(/["']resource:\/\/[^"']+["']/g);
         if (
@@ -552,6 +701,18 @@ function parseCodeFile(fileUri) {
           addCodeReference(convertToCodeURI(url), fileUri);
         }
 
+        // This handles `import` lines which may be multi-line.
+        // We have an ESLint rule, `import/no-unassigned-import` which prevents
+        // using bare `import "foo.js"`, so we don't need to handle that case
+        // here.
+        match = line.match(/from\W*['"](.*?)['"]/);
+        if (match?.[1]) {
+          let url = match[1];
+          url = Services.io.newURI(url, null, baseUri || fileUri).spec;
+          url = convertToCodeURI(url);
+          addCodeReference(url, fileUri);
+        }
+
         if (isDevtools) {
           let rules = [
             ["devtools/client/locales", "chrome://devtools/locale"],
@@ -569,7 +730,7 @@ function parseCodeFile(fileUri) {
             for (let rule of rules) {
               if (path.startsWith(rule[0] + "/")) {
                 path = path.replace(rule[0], rule[1]);
-                if (!/\.(properties|js|jsm|json|css)$/.test(path)) {
+                if (!/\.(properties|js|jsm|mjs|json|css)$/.test(path)) {
                   path += ".js";
                 }
                 addCodeReference(path, fileUri);
@@ -583,10 +744,13 @@ function parseCodeFile(fileUri) {
             let url = match[1];
             url = Services.io.newURI(url, null, baseUri || fileUri).spec;
             url = convertToCodeURI(url);
-            if (!/\.(properties|js|jsm|json|css)$/.test(url)) {
+            if (!/\.(properties|js|jsm|mjs|json|css)$/.test(url)) {
               url += ".js";
             }
-            if (url.startsWith("resource://")) {
+            if (
+              url.startsWith("resource://") ||
+              url.startsWith("moz-src:///")
+            ) {
               addCodeReference(url, fileUri);
             } else {
               // if we end up with a chrome:// url here, it's likely because
@@ -622,7 +786,7 @@ function parseCodeFile(fileUri) {
         if (
           isDevtools &&
           line.includes("require(") &&
-          !/\.(properties|js|jsm|json|css)$/.test(url)
+          !/\.(properties|js|jsm|mjs|json|css)$/.test(url)
         ) {
           url += ".js";
         }
@@ -636,13 +800,22 @@ function parseCodeFile(fileUri) {
 function convertToCodeURI(fileUri) {
   let baseUri = fileUri;
   let path = "";
-  while (true) {
+  while (baseUri) {
     let slashPos = baseUri.lastIndexOf("/", baseUri.length - 2);
     if (slashPos <= 0) {
       // File not accessible from chrome protocol, try resource://
       for (let res of gResourceMap) {
         if (fileUri.startsWith(res[1])) {
-          return fileUri.replace(res[1], "resource://" + res[0] + "/");
+          let resourceUriString = fileUri.replace(
+            res[1],
+            `resource://${res[0]}/`
+          );
+          // If inside moz-src, treat as moz-src url.
+          resourceUriString = resourceUriString.replace(
+            /^resource:\/\/gre\/moz-src\//,
+            "moz-src:///"
+          );
+          return resourceUriString;
         }
       }
       // Give up and return the original URL.
@@ -654,31 +827,16 @@ function convertToCodeURI(fileUri) {
       return gChromeMap.get(baseUri) + path;
     }
   }
+  throw new Error(`Unparsable URI: ${fileUri}`);
 }
 
-function chromeFileExists(aURI) {
-  let available = 0;
+async function chromeFileExists(aURI) {
   try {
-    let channel = NetUtil.newChannel({
-      uri: aURI,
-      loadUsingSystemPrincipal: true,
-    });
-    let stream = channel.open();
-    let sstream = Cc["@mozilla.org/scriptableinputstream;1"].createInstance(
-      Ci.nsIScriptableInputStream
-    );
-    sstream.init(stream);
-    available = sstream.available();
-    sstream.close();
+    return await PerfTestHelpers.checkURIExists(aURI);
   } catch (e) {
-    if (
-      e.result != Cr.NS_ERROR_FILE_NOT_FOUND &&
-      e.result != Cr.NS_ERROR_NOT_AVAILABLE
-    ) {
-      todo(false, "Failed to check if " + aURI + "exists: " + e);
-    }
+    todo(false, `Failed to check if ${aURI} exists: ${e}`);
+    return false;
   }
-  return available > 0;
 }
 
 function findChromeUrlsFromArray(array, prefix) {
@@ -717,6 +875,7 @@ function findChromeUrlsFromArray(array, prefix) {
     // Only keep strings that look like real chrome or resource urls.
     if (
       /chrome:\/\/[a-zA-Z09-]+\/(content|skin|locale)\//.test(string) ||
+      /moz-src:\/\/\/\w+/.test(string) ||
       /resource:\/\/[a-zA-Z09-]*\/.*\.[a-z]+/.test(string)
     ) {
       gReferencesFromCode.set(string, null);
@@ -725,24 +884,23 @@ function findChromeUrlsFromArray(array, prefix) {
 }
 
 add_task(async function checkAllTheFiles() {
-  let libxulPath = OS.Constants.Path.libxul;
-  if (AppConstants.platform != "macosx") {
-    libxulPath = OS.Constants.Path.libDir + "/" + libxulPath;
-  }
-  let libxul = await OS.File.read(libxulPath);
+  TestUtils.assertPackagedBuild();
+
+  const libxul = await IOUtils.read(PathUtils.xulLibraryPath);
   findChromeUrlsFromArray(libxul, "chrome://");
   findChromeUrlsFromArray(libxul, "resource://");
+  findChromeUrlsFromArray(libxul, "moz-src://");
   // Handle NS_LITERAL_STRING.
   let uint16 = new Uint16Array(libxul.buffer);
   findChromeUrlsFromArray(uint16, "chrome://");
   findChromeUrlsFromArray(uint16, "resource://");
+  findChromeUrlsFromArray(uint16, "moz-src://");
 
   const kCodeExtensions = [
-    ".xul",
     ".xml",
     ".xsl",
+    ".mjs",
     ".js",
-    ".jsm",
     ".json",
     ".html",
     ".xhtml",
@@ -786,10 +944,10 @@ add_task(async function checkAllTheFiles() {
   });
 
   // Wait for all manifest to be parsed
-  await throttledMapPromises(manifestURIs, parseManifest);
+  await PerfTestHelpers.throttledMapPromises(manifestURIs, parseManifest);
 
-  for (let jsm of Components.manager.getComponentJSMs()) {
-    gReferencesFromCode.set(jsm, null);
+  for (let esModule of Components.manager.getComponentESModules()) {
+    gReferencesFromCode.set(esModule, null);
   }
 
   // manifest.json is a common name, it is used for WebExtension manifests
@@ -816,15 +974,18 @@ add_task(async function checkAllTheFiles() {
   }
 
   // Wait for all the files to have actually loaded:
-  await throttledMapPromises(allPromises, ([task, uri]) => task(uri));
+  await PerfTestHelpers.throttledMapPromises(allPromises, ([task, uri]) =>
+    task(uri)
+  );
 
   // Keep only chrome:// files, and filter out either the devtools paths or
   // the non-devtools paths:
   let devtoolsPrefixes = [
     "chrome://devtools",
+    "moz-src:///devtools/",
     "resource://devtools/",
-    "resource://devtools-client-jsonview/",
-    "resource://devtools-client-shared/",
+    "resource://devtools-shared-images/",
+    "resource://devtools-highlighter-styles/",
     "resource://app/modules/devtools",
     "resource://gre/modules/devtools",
     "resource://app/localization/en-US/startup/aboutDevTools.ftl",
@@ -836,7 +997,9 @@ add_task(async function checkAllTheFiles() {
   for (let uri of uris) {
     uri = convertToCodeURI(uri.spec);
     if (
-      (uri.startsWith("chrome://") || uri.startsWith("resource://")) &&
+      (uri.startsWith("chrome://") ||
+        uri.startsWith("resource://") ||
+        uri.startsWith("moz-src:///")) &&
       isDevtools == hasDevtoolsPrefix(uri)
     ) {
       chromeFiles.push(uri);
@@ -863,7 +1026,8 @@ add_task(async function checkAllTheFiles() {
         if (isDevtools) {
           if (
             ref.startsWith("resource://app/components/") ||
-            (file.startsWith("chrome://") && ref.startsWith("resource://"))
+            (file.startsWith("chrome://") &&
+              (ref.startsWith("resource://") || ref.startsWith("moz-src://")))
           ) {
             return false;
           }
@@ -873,8 +1037,8 @@ add_task(async function checkAllTheFiles() {
           let refType = gReferencesFromCode.get(ref);
           if (
             refType === null || // unconditionally referenced
-            refType == "whitelist" ||
-            refType == "whitelist-direct"
+            refType == "allowlist" ||
+            refType == "allowlist-direct"
           ) {
             return false;
           }
@@ -886,27 +1050,24 @@ add_task(async function checkAllTheFiles() {
 
   let unreferencedFiles = chromeFiles;
 
-  let removeReferenced = useWhitelist => {
+  let removeReferenced = useAllowlist => {
     let foundReference = false;
     unreferencedFiles = unreferencedFiles.filter(f => {
       let rv = isUnreferenced(f);
       if (rv && f.startsWith("resource://app/")) {
         rv = isUnreferenced(f.replace("resource://app/", "resource:///"));
       }
-      if (rv && /^resource:\/\/(?:app|gre)\/components\/[^/]+\.js$/.test(f)) {
-        rv = !gComponentsSet.has(f.replace(/.*\//, ""));
-      }
       if (!rv) {
         foundReference = true;
-        if (useWhitelist) {
+        if (useAllowlist) {
           info(
-            "indirectly whitelisted file: " +
+            "indirectly allowlisted file: " +
               f +
               " used from " +
               listCodeReferences(gReferencesFromCode.get(f))
           );
         }
-        gReferencesFromCode.set(f, useWhitelist ? "whitelist" : null);
+        gReferencesFromCode.set(f, useAllowlist ? "allowlist" : null);
       }
       return rv;
     });
@@ -917,17 +1078,17 @@ add_task(async function checkAllTheFiles() {
     // As long as removeReferenced returns true, some files have been marked
     // as referenced, so we need to run it again.
   }
-  // Marked as referenced the files that have been explicitly whitelisted.
+  // Marked as referenced the files that have been explicitly allowed.
   unreferencedFiles = unreferencedFiles.filter(file => {
-    if (whitelist.has(file)) {
-      whitelist.delete(file);
-      gReferencesFromCode.set(file, "whitelist-direct");
+    if (allowlist.has(file)) {
+      allowlist.delete(file);
+      gReferencesFromCode.set(file, "allowlist-direct");
       return false;
     }
     return true;
   });
   // Run the process again, this time when more files are marked as referenced,
-  // it's a consequence of the whitelist.
+  // it's a consequence of the allowlist.
   while (removeReferenced(true)) {
     // As long as removeReferenced returns true, we need to run it again.
   }
@@ -960,11 +1121,11 @@ add_task(async function checkAllTheFiles() {
     }
   }
 
-  for (let file of whitelist) {
-    if (ignorableWhitelist.has(file)) {
-      info("ignored unused whitelist entry: " + file);
+  for (let file of allowlist) {
+    if (ignorableAllowlist.has(file)) {
+      info("ignored unused allowlist entry: " + file);
     } else {
-      ok(false, "unused whitelist entry: " + file);
+      ok(false, "unused allowlist entry: " + file);
     }
   }
 
@@ -976,8 +1137,10 @@ add_task(async function checkAllTheFiles() {
     }
 
     if (
-      (file.startsWith("chrome://") || file.startsWith("resource://")) &&
-      !chromeFileExists(file)
+      (file.startsWith("chrome://") ||
+        file.startsWith("resource://") ||
+        file.startsWith("moz-src:///")) &&
+      !(await chromeFileExists(file))
     ) {
       // Ignore chrome prefixes that have been automatically expanded.
       let pathParts =
@@ -987,7 +1150,7 @@ add_task(async function checkAllTheFiles() {
         continue;
       }
 
-      // TODO: bug 1349010 - add a whitelist and make this reliable enough
+      // TODO: bug 1349010 - add a allowlist and make this reliable enough
       // that we could make the test fail when this catches something new.
       let refList = listCodeReferences(refs);
       let msg = "missing file: " + file;
