@@ -8,26 +8,23 @@
 # It is defined inline because this was easiest to make test archive
 # generation faster.
 
-from __future__ import absolute_import, print_function, unicode_literals
-
 import argparse
 import itertools
 import os
 import sys
 import time
 
+import buildconfig
+import mozpack.path as mozpath
 from manifestparser import TestManifest
-from reftest import ReftestManifest
-
-from mozbuild.util import ensureParentDir
 from mozpack.archive import create_tar_gz_from_files
 from mozpack.copier import FileRegistry
 from mozpack.files import ExistingFile, FileFinder
 from mozpack.manifests import InstallManifest
 from mozpack.mozjar import JarWriter
-import mozpack.path as mozpath
+from reftest import ReftestManifest
 
-import buildconfig
+from mozbuild.dirutils import ensureParentDir
 
 STAGE = mozpath.join(buildconfig.topobjdir, "dist", "test-stage")
 
@@ -35,6 +32,7 @@ TEST_HARNESS_BINS = [
     "BadCertAndPinningServer",
     "DelegatedCredentialsServer",
     "EncryptedClientHelloServer",
+    "FaultyServer",
     "GenerateOCSPResponse",
     "OCSPStaplingServer",
     "SanctionsTestServer",
@@ -43,21 +41,22 @@ TEST_HARNESS_BINS = [
     "crashinject",
     "geckodriver",
     "http3server",
+    "content_analysis_sdk_agent",
     "minidumpwriter",
     "pk12util",
     "screenshot",
     "screentopng",
     "ssltunnel",
+    "test_stub_installer",
     "xpcshell",
+    "plugin-container",
 ]
 
 TEST_HARNESS_DLLS = ["crashinjectdll", "mozglue"]
 
-GMP_TEST_PLUGIN_DIRS = [
-    "gmp-clearkey/**",
-    "gmp-fake/**",
-    "gmp-fakeopenh264/**",
-]
+TRAIN_HOP_DLLS = ["xul", "nss3", "gkcodecs", "lgpllibs"]
+
+GMP_TEST_PLUGIN_DIRS = ["gmp-fake/**", "gmp-fakeopenh264/**"]
 
 # These entries will be used by artifact builds to re-construct an
 # objdir with the appropriate generated support files.
@@ -99,49 +98,33 @@ ARCHIVE_FILES = {
                 "jittest/**",  # To make the ignore checker happy
                 "perftests/**",
                 "fuzztest/**",
+                "trainhop/**",
             ],
         },
-        {
-            "source": buildconfig.topobjdir,
-            "base": "_tests",
-            "pattern": "modules/**",
-        },
+        {"source": buildconfig.topobjdir, "base": "_tests", "pattern": "modules/**"},
         {
             "source": buildconfig.topsrcdir,
             "base": "testing/marionette",
-            "patterns": [
-                "client/**",
-                "harness/**",
-                "mach_test_package_commands.py",
-            ],
+            "patterns": ["client/**", "harness/**", "mach_test_package_commands.py"],
             "dest": "marionette",
-            "ignore": [
-                "client/docs",
-                "harness/marionette_harness/tests",
-            ],
+            "ignore": ["client/docs", "harness/marionette_harness/tests"],
         },
         {
             "source": buildconfig.topsrcdir,
             "base": "",
             "manifests": [
-                "testing/marionette/harness/marionette_harness/tests/unit-tests.ini",
+                "testing/marionette/harness/marionette_harness/tests/unit-tests.toml"
             ],
             # We also need the manifests and harness_unit tests
             "pattern": "testing/marionette/harness/marionette_harness/tests/**",
             "dest": "marionette/tests",
         },
-        {
-            "source": buildconfig.topobjdir,
-            "base": "_tests",
-            "pattern": "mozbase/**",
-        },
+        {"source": buildconfig.topobjdir, "base": "_tests", "pattern": "mozbase/**"},
         {
             "source": buildconfig.topsrcdir,
             "base": "testing",
             "pattern": "firefox-ui/**",
-            "ignore": [
-                "firefox-ui/tests",
-            ],
+            "ignore": ["firefox-ui/tests"],
         },
         {
             "source": buildconfig.topsrcdir,
@@ -155,11 +138,7 @@ ARCHIVE_FILES = {
             "pattern": "/**",
             "dest": "telemetry/marionette",
         },
-        {
-            "source": buildconfig.topsrcdir,
-            "base": "testing",
-            "pattern": "tps/**",
-        },
+        {"source": buildconfig.topsrcdir, "base": "testing", "pattern": "tps/**"},
         {
             "source": buildconfig.topsrcdir,
             "base": "services/sync/",
@@ -189,6 +168,12 @@ ARCHIVE_FILES = {
             "pattern": "**",
             "dest": "tools/mozterm",
         },
+        {
+            "source": buildconfig.topsrcdir,
+            "base": "xpcom/geckoprocesstypes_generator",
+            "pattern": "**",
+            "dest": "tools/geckoprocesstypes_generator",
+        },
         {
             "source": buildconfig.topsrcdir,
             "base": "third_party/python/six",
@@ -201,11 +186,7 @@ ARCHIVE_FILES = {
             "pattern": "**",
             "dest": "tools/distro",
         },
-        {
-            "source": buildconfig.topobjdir,
-            "base": "",
-            "pattern": "mozinfo.json",
-        },
+        {"source": buildconfig.topobjdir, "base": "", "pattern": "mozinfo.json"},
         {
             "source": buildconfig.topobjdir,
             "base": "dist/bin",
@@ -233,18 +214,13 @@ ARCHIVE_FILES = {
         {
             "source": buildconfig.topobjdir,
             "base": "dist/bin",
-            "patterns": [
-                "dmd.py",
-                "fix_stacks.py",
-            ],
+            "patterns": ["dmd.py", "fix_stacks.py"],
             "dest": "bin",
         },
         {
             "source": buildconfig.topobjdir,
             "base": "dist/bin/components",
-            "patterns": [
-                "httpd.js",
-            ],
+            "patterns": ["httpd.sys.mjs"],
             "dest": "bin/components",
         },
         {
@@ -255,11 +231,7 @@ ARCHIVE_FILES = {
         },
     ],
     "cppunittest": [
-        {
-            "source": STAGE,
-            "base": "",
-            "pattern": "cppunittest/**",
-        },
+        {"source": STAGE, "base": "", "pattern": "cppunittest/**"},
         # We don't ship these files if startup cache is disabled, which is
         # rare. But it shouldn't matter for test archives.
         {
@@ -283,7 +255,7 @@ ARCHIVE_FILES = {
         {
             "source": buildconfig.topsrcdir,
             "base": "testing",
-            "pattern": "cppunittest.ini",
+            "pattern": "cppunittest.toml",
             "dest": "cppunittest",
         },
         {
@@ -293,13 +265,7 @@ ARCHIVE_FILES = {
             "dest": "cppunittest",
         },
     ],
-    "gtest": [
-        {
-            "source": STAGE,
-            "base": "",
-            "pattern": "gtest/**",
-        },
-    ],
+    "gtest": [{"source": STAGE, "base": "", "pattern": "gtest/**"}],
     "mochitest": [
         OBJDIR_TEST_FILES["mochitest"],
         {
@@ -307,11 +273,7 @@ ARCHIVE_FILES = {
             "base": "_tests/testing",
             "pattern": "mochitest/**",
         },
-        {
-            "source": STAGE,
-            "base": "",
-            "pattern": "mochitest/**",
-        },
+        {"source": STAGE, "base": "", "pattern": "mochitest/**"},
         {
             "source": buildconfig.topobjdir,
             "base": "",
@@ -330,6 +292,19 @@ ARCHIVE_FILES = {
             "pattern": "specialpowers/**",
             "dest": "mochitest/extensions",
         },
+        # Needed by Windows a11y browser tests.
+        {
+            "source": buildconfig.topobjdir,
+            "base": "accessible/interfaces/ia2",
+            "pattern": "IA2Typelib.tlb",
+            "dest": "mochitest",
+        },
+        {
+            "source": buildconfig.topobjdir,
+            "base": "dist/bin",
+            "pattern": "IA2Marshal.dll",
+            "dest": "mochitest",
+        },
     ],
     "mozharness": [
         {
@@ -340,7 +315,7 @@ ARCHIVE_FILES = {
         {
             "source": buildconfig.topsrcdir,
             "base": "",
-            "pattern": "third_party/python/virtualenv/**",
+            "pattern": "third_party/python/_venv/**",
         },
         {
             "source": buildconfig.topsrcdir,
@@ -357,28 +332,60 @@ ARCHIVE_FILES = {
             "base": "testing/mozbase/mozinfo",
             "pattern": "mozinfo/**",
         },
+        {
+            "source": buildconfig.topsrcdir,
+            "base": "testing/mozbase/mozlog",
+            "pattern": "mozlog/**",
+        },
+        {
+            "source": buildconfig.topsrcdir,
+            "base": "python/mozterm",
+            "pattern": "mozterm/**",
+        },
         {
             "source": buildconfig.topsrcdir,
             "base": "testing/mozbase/mozprocess",
             "pattern": "mozprocess/**",
         },
+        {
+            "source": buildconfig.topsrcdir,
+            "base": "testing/mozbase/mozsystemmonitor",
+            "pattern": "mozsystemmonitor/**",
+        },
         {
             "source": buildconfig.topsrcdir,
             "base": "third_party/python/six",
             "pattern": "six.py",
         },
+        {
+            "source": buildconfig.topsrcdir,
+            "base": "third_party/python/toml",
+            "pattern": "**",
+        },
+        {
+            "source": buildconfig.topsrcdir,
+            "base": "third_party/python/tomlkit",
+            "pattern": "**",
+        },
         {
             "source": buildconfig.topsrcdir,
             "base": "third_party/python/distro",
-            "pattern": "distro.py",
+            "pattern": "distro/**",
+        },
+        {
+            "source": buildconfig.topsrcdir,
+            "base": "third_party/python/packaging",
+            "pattern": "**",
+        },
+        {
+            "source": buildconfig.topsrcdir,
+            "base": "python/mozbuild/mozbuild/action",
+            "pattern": "tooltool.py",
+            "dest": "external_tools",
         },
     ],
     "reftest": [
-        {
-            "source": buildconfig.topobjdir,
-            "base": "_tests",
-            "pattern": "reftest/**",
-        },
+        {"source": buildconfig.topobjdir, "base": "_tests", "pattern": "reftest/**"},
         {
             "source": buildconfig.topobjdir,
             "base": "",
@@ -409,11 +416,7 @@ ARCHIVE_FILES = {
         },
     ],
     "talos": [
-        {
-            "source": buildconfig.topsrcdir,
-            "base": "testing",
-            "pattern": "talos/**",
-        },
+        {"source": buildconfig.topsrcdir, "base": "testing", "pattern": "talos/**"},
         {
             "source": buildconfig.topsrcdir,
             "base": "testing/profiles",
@@ -428,39 +431,28 @@ ARCHIVE_FILES = {
         },
     ],
     "perftests": [
-        {
-            "source": buildconfig.topsrcdir,
-            "pattern": "testing/mozbase/**",
-        },
-        {
-            "source": buildconfig.topsrcdir,
-            "pattern": "testing/condprofile/**",
-        },
-        {
-            "source": buildconfig.topsrcdir,
-            "pattern": "third_party/python/**",
-        },
-        {
-            "source": buildconfig.topsrcdir,
-            "pattern": "tools/lint/eslint/**",
-        },
+        {"source": buildconfig.topsrcdir, "pattern": "testing/mozbase/**"},
+        {"source": buildconfig.topsrcdir, "pattern": "testing/condprofile/**"},
+        {"source": buildconfig.topsrcdir, "pattern": "testing/performance/**"},
+        {"source": buildconfig.topsrcdir, "pattern": "third_party/python/**"},
+        {"source": buildconfig.topsrcdir, "pattern": "tools/lint/eslint/**"},
         {"source": buildconfig.topsrcdir, "pattern": "**/perftest_*.js"},
         {"source": buildconfig.topsrcdir, "pattern": "**/hooks_*py"},
         {"source": buildconfig.topsrcdir, "pattern": "build/autoconf/**"},
         {"source": buildconfig.topsrcdir, "pattern": "build/moz.configure/**"},
         {"source": buildconfig.topsrcdir, "pattern": "python/**"},
-        {"source": buildconfig.topsrcdir, "pattern": "build/mach_bootstrap.py"},
+        {"source": buildconfig.topsrcdir, "pattern": "build/mach_initialize.py"},
         {
             "source": buildconfig.topsrcdir,
-            "pattern": "build/build_virtualenv_packages.txt",
+            "pattern": "python/sites/build.txt",
         },
         {
             "source": buildconfig.topsrcdir,
-            "pattern": "build/common_virtualenv_packages.txt",
+            "pattern": "python/sites/common.txt",
         },
         {
             "source": buildconfig.topsrcdir,
-            "pattern": "build/mach_virtualenv_packages.txt",
+            "pattern": "python/sites/mach.txt",
         },
         {"source": buildconfig.topsrcdir, "pattern": "mach/**"},
         {
@@ -483,11 +475,13 @@ ARCHIVE_FILES = {
                 "chrome/**",
                 "chrome.manifest",
                 "components/**",
+                "content_analysis_sdk_agent",
                 "http3server",
                 "*.ini",
+                "*.toml",
                 "localization/**",
                 "modules/**",
-                "update.locale",
+                "default.locale",
                 "greprefs.js",
             ],
             "dest": "bin",
@@ -537,11 +531,7 @@ ARCHIVE_FILES = {
         },
     ],
     "raptor": [
-        {
-            "source": buildconfig.topsrcdir,
-            "base": "testing",
-            "pattern": "raptor/**",
-        },
+        {"source": buildconfig.topsrcdir, "base": "testing", "pattern": "raptor/**"},
         {
             "source": buildconfig.topsrcdir,
             "base": "testing/profiles",
@@ -556,11 +546,7 @@ ARCHIVE_FILES = {
         },
     ],
     "awsy": [
-        {
-            "source": buildconfig.topsrcdir,
-            "base": "testing",
-            "pattern": "awsy/**",
-        },
+        {"source": buildconfig.topsrcdir, "base": "testing", "pattern": "awsy/**"}
     ],
     "web-platform": [
         {
@@ -577,10 +563,7 @@ ARCHIVE_FILES = {
             "source": buildconfig.topsrcdir,
             "base": "testing",
             "pattern": "web-platform/tests/**",
-            "ignore": [
-                "web-platform/tests/tools/wptserve",
-                "web-platform/tests/tools/wpt_third_party",
-            ],
+            "ignore": ["web-platform/tests/tools/wpt_third_party"],
         },
         {
             "source": buildconfig.topobjdir,
@@ -604,7 +587,8 @@ ARCHIVE_FILES = {
                 "mach_test_package_commands.py",
                 "moz-http2/**",
                 "node-http2/**",
-                "node-ip/**",
+                "node_ip/**",
+                "node-ws/**",
                 "dns-packet/**",
                 "remotexpcshelltests.py",
                 "runxpcshelltests.py",
@@ -613,11 +597,7 @@ ARCHIVE_FILES = {
             ],
             "dest": "xpcshell",
         },
-        {
-            "source": STAGE,
-            "base": "",
-            "pattern": "xpcshell/**",
-        },
+        {"source": STAGE, "base": "", "pattern": "xpcshell/**"},
         {
             "source": buildconfig.topobjdir,
             "base": "",
@@ -664,18 +644,9 @@ ARCHIVE_FILES = {
             "dest": "updater-dep",
         },
     ],
-    "jsreftest": [
-        {
-            "source": STAGE,
-            "base": "",
-            "pattern": "jsreftest/**",
-        },
-    ],
+    "jsreftest": [{"source": STAGE, "base": "", "pattern": "jsreftest/**"}],
     "fuzztest": [
-        {
-            "source": buildconfig.topsrcdir,
-            "pattern": "tools/fuzzing/smoke/**",
-        },
+        {"source": buildconfig.topsrcdir, "pattern": "tools/fuzzing/smoke/**"}
     ],
     "jittest": [
         {
@@ -715,6 +686,26 @@ ARCHIVE_FILES = {
             "dest": "jit-test",
         },
     ],
+    "trainhop": [
+        {
+            "source": buildconfig.topobjdir,
+            "base": "dist/bin",
+            "patterns": [
+                "%s%s" % (f, buildconfig.substs["BIN_SUFFIX"])
+                for f in TEST_HARNESS_BINS
+            ]
+            + [
+                "%s%s%s"
+                % (
+                    buildconfig.substs["DLL_PREFIX"],
+                    f,
+                    buildconfig.substs["DLL_SUFFIX"],
+                )
+                for f in TRAIN_HOP_DLLS
+            ],
+            "dest": "bin",
+        },
+    ],
 }
 
 if buildconfig.substs.get("MOZ_CODE_COVERAGE"):
@@ -722,15 +713,15 @@ if buildconfig.substs.get("MOZ_CODE_COVERAGE"):
         {
             "source": buildconfig.topsrcdir,
             "base": "python/mozbuild/",
-            "patterns": [
-                "mozpack/**",
-                "mozbuild/codecoverage/**",
-            ],
+            "patterns": ["mozpack/**", "mozbuild/codecoverage/**"],
         }
     )
 
 
-if buildconfig.substs.get("MOZ_ASAN") and buildconfig.substs.get("CLANG_CL"):
+if (
+    buildconfig.substs.get("MOZ_ASAN")
+    and buildconfig.substs.get("CC_TYPE") == "clang-cl"
+):
     asan_dll = {
         "source": buildconfig.topobjdir,
         "base": "dist/bin",
@@ -751,7 +742,7 @@ if buildconfig.substs.get("commtopsrcdir"):
     marionette_comm = {
         "source": commtopsrcdir,
         "base": "",
-        "manifest": "testing/marionette/unit-tests.ini",
+        "manifest": "testing/marionette/unit-tests.toml",
         "dest": "marionette/tests/comm",
     }
     ARCHIVE_FILES["common"].append(marionette_comm)
@@ -817,11 +808,7 @@ def find_files(archive):
             if patterns:
                 entry["patterns"] = patterns
                 extra_entries.append(entry)
-        entry = {
-            "source": buildconfig.topobjdir,
-            "base": "_tests",
-            "patterns": [],
-        }
+        entry = {"source": buildconfig.topobjdir, "base": "_tests", "patterns": []}
         for path in set(generated_harness_files) - packaged_paths:
             entry["patterns"].append(path[len("_tests") + 1 :])
         extra_entries.append(entry)
@@ -842,16 +829,10 @@ def find_files(archive):
             manifests.append(manifest)
         if manifests:
             dirs = find_manifest_dirs(os.path.join(source, base), manifests)
-            patterns.extend({"{}/**".format(d) for d in dirs})
+            patterns.extend({f"{d}/**" for d in dirs})
 
         ignore = list(entry.get("ignore", []))
-        ignore.extend(
-            [
-                "**/.flake8",
-                "**/.mkdir.done",
-                "**/*.pyc",
-            ]
-        )
+        ignore.extend(["**/.flake8", "**/.mkdir.done", "**/*.pyc"])
 
         if archive not in ("common", "updater-dep") and base.startswith("_tests"):
             # We may have generated_harness_files to exclude from this entry.
@@ -859,10 +840,7 @@ def find_files(archive):
                 if path.startswith(base):
                     ignore.append(path[len(base) + 1 :])
 
-        common_kwargs = {
-            "find_dotfiles": True,
-            "ignore": ignore,
-        }
+        common_kwargs = {"find_dotfiles": True, "ignore": ignore}
 
         finder = FileFinder(os.path.join(source, base), **common_kwargs)
 
@@ -883,7 +861,7 @@ def find_manifest_dirs(topsrcdir, manifests):
     for p in manifests:
         p = os.path.join(topsrcdir, p)
 
-        if p.endswith(".ini"):
+        if p.endswith(".ini") or p.endswith(".toml"):
             test_manifest = TestManifest()
             test_manifest.read(p)
             dirs |= set([os.path.dirname(m) for m in test_manifest.manifests()])
@@ -895,9 +873,7 @@ def find_manifest_dirs(topsrcdir, manifests):
 
         else:
             raise Exception(
-                '"{}" is not a supported manifest format.'.format(
-                    os.path.splitext(p)[1]
-                )
+                f'"{os.path.splitext(p)[1]}" is not a supported manifest format.'
             )
 
     dirs = {mozpath.normpath(d[len(topsrcdir) :]).lstrip("/") for d in dirs}
@@ -931,7 +907,7 @@ def main(argv):
         raise Exception("expected tar.gz or zip output file")
 
     file_count = 0
-    t_start = time.time()
+    t_start = time.monotonic()
     ensureParentDir(out_file)
     res = find_files(args.archive)
     with open(out_file, "wb") as fh:
@@ -953,7 +929,7 @@ def main(argv):
         else:
             raise Exception("unhandled file extension: %s" % out_file)
 
-    duration = time.time() - t_start
+    duration = time.monotonic() - t_start
     zip_size = os.path.getsize(args.outputfile)
     basename = os.path.basename(args.outputfile)
     print(
