@@ -13,6 +13,7 @@
 #include "nsProxyRelease.h"
 #include "nsServiceManagerUtils.h"
 #include "nsITimer.h"
+#include "nsThreadManager.h"
 #include "nsThreadUtils.h"
 #include "nsPIDOMWindow.h"
 #include "nsIObserverService.h"
@@ -35,12 +36,21 @@
 #include "mozilla/RDDProcessManager.h"
 #include "mozilla/ResultExtensions.h"
 #include "mozilla/Services.h"
-#include "mozilla/Telemetry.h"
+#include "mozilla/glean/XpcomMetrics.h"
 #include "mozilla/UniquePtrExtensions.h"
 #include "mozilla/dom/MemoryReportTypes.h"
 #include "mozilla/dom/ContentParent.h"
 #include "mozilla/gfx/GPUProcessManager.h"
+#include "mozilla/ipc/UtilityProcessManager.h"
 #include "mozilla/ipc/FileDescriptorUtils.h"
+#ifdef MOZ_PHC
+#  include "PHC.h"
+#endif
+
+#ifdef MOZ_WIDGET_ANDROID
+#  include "mozilla/java/GeckoAppShellWrappers.h"
+#  include "mozilla/jni/Utils.h"
+#endif
 
 #ifdef XP_WIN
 #  include "mozilla/MemoryInfo.h"
@@ -54,13 +64,9 @@
 #endif
 
 using namespace mozilla;
+using namespace mozilla::ipc;
 using namespace dom;
 
-#if defined(MOZ_MEMORY)
-#  define HAVE_JEMALLOC_STATS 1
-#  include "mozmemory.h"
-#endif  // MOZ_MEMORY
-
 #if defined(XP_LINUX)
 
 #  include "mozilla/MemoryMapping.h"
@@ -188,6 +194,7 @@ using namespace dom;
   return NS_ERROR_FAILURE;
 #  endif
   int mib[] = {
+      // clang-format off
     CTL_KERN,
     KERN_PROC,
     KERN_PROC_PID,
@@ -196,6 +203,7 @@ using namespace dom;
     sizeof(KINFO_PROC),
     1,
 #  endif
+      // clang-format on
   };
   u_int miblen = sizeof(mib) / sizeof(mib[0]);
   size_t size = sizeof(KINFO_PROC);
@@ -409,7 +417,7 @@ static void XMappingIter(int64_t& aVsize, int64_t& aResident,
                                                                 bool aDoPurge) {
 #  ifdef HAVE_JEMALLOC_STATS
   if (aDoPurge) {
-    Telemetry::AutoTimer<Telemetry::MEMORY_FREE_PURGED_PAGES_MS> timer;
+    auto timer = glean::memory::free_purged_pages.Measure();
     jemalloc_purge_freed_pages();
   }
 #  endif
@@ -441,6 +449,10 @@ static bool InSharedRegion(mach_vm_address_t aAddr, cpu_type_t aType) {
       base = SHARED_REGION_BASE_ARM;
       size = SHARED_REGION_SIZE_ARM;
       break;
+    case CPU_TYPE_ARM64:
+      base = SHARED_REGION_BASE_ARM64;
+      size = SHARED_REGION_SIZE_ARM64;
+      break;
     case CPU_TYPE_I386:
       base = SHARED_REGION_BASE_I386;
       size = SHARED_REGION_SIZE_I386;
@@ -471,15 +483,16 @@ static bool InSharedRegion(mach_vm_address_t aAddr, cpu_type_t aType) {
   // Roughly based on libtop_update_vm_regions in
   // http://www.opensource.apple.com/source/top/top-100.1.2/libtop.c
   size_t privatePages = 0;
-  mach_vm_size_t size = 0;
-  for (mach_vm_address_t addr = MACH_VM_MIN_ADDRESS;; addr += size) {
-    vm_region_top_info_data_t info;
-    mach_msg_type_number_t infoCount = VM_REGION_TOP_INFO_COUNT;
-    mach_port_t objectName;
+  mach_vm_size_t topSize = 0;
+  for (mach_vm_address_t addr = MACH_VM_MIN_ADDRESS;; addr += topSize) {
+    vm_region_top_info_data_t topInfo;
+    mach_msg_type_number_t topInfoCount = VM_REGION_TOP_INFO_COUNT;
+    mach_port_t topObjectName;
 
     kern_return_t kr = mach_vm_region(
-        aPort ? aPort : mach_task_self(), &addr, &size, VM_REGION_TOP_INFO,
-        reinterpret_cast<vm_region_info_t>(&info), &infoCount, &objectName);
+        aPort ? aPort : mach_task_self(), &addr, &topSize, VM_REGION_TOP_INFO,
+        reinterpret_cast<vm_region_info_t>(&topInfo), &topInfoCount,
+        &topObjectName);
     if (kr == KERN_INVALID_ADDRESS) {
       // Done iterating VM regions.
       break;
@@ -487,26 +500,51 @@ static bool InSharedRegion(mach_vm_address_t aAddr, cpu_type_t aType) {
       return NS_ERROR_FAILURE;
     }
 
-    if (InSharedRegion(addr, cpu_type) && info.share_mode != SM_PRIVATE) {
+    if (InSharedRegion(addr, cpu_type) && topInfo.share_mode != SM_PRIVATE) {
       continue;
     }
 
-    switch (info.share_mode) {
+    switch (topInfo.share_mode) {
       case SM_LARGE_PAGE:
         // NB: Large pages are not shareable and always resident.
       case SM_PRIVATE:
-        privatePages += info.private_pages_resident;
-        privatePages += info.shared_pages_resident;
+        privatePages += topInfo.private_pages_resident;
+        privatePages += topInfo.shared_pages_resident;
         break;
       case SM_COW:
-        privatePages += info.private_pages_resident;
-        if (info.ref_count == 1) {
+        privatePages += topInfo.private_pages_resident;
+        if (topInfo.ref_count == 1) {
           // Treat copy-on-write pages as private if they only have one
           // reference.
-          privatePages += info.shared_pages_resident;
+          privatePages += topInfo.shared_pages_resident;
         }
         break;
-      case SM_SHARED:
+      case SM_SHARED: {
+        // Using mprotect() or similar to protect a page in the middle of a
+        // mapping can create aliased mappings. They look like shared mappings
+        // to the VM_REGION_TOP_INFO interface, so re-check with
+        // VM_REGION_EXTENDED_INFO.
+
+        mach_vm_size_t exSize = 0;
+        vm_region_extended_info_data_t exInfo;
+        mach_msg_type_number_t exInfoCount = VM_REGION_EXTENDED_INFO_COUNT;
+        mach_port_t exObjectName;
+        kr = mach_vm_region(aPort ? aPort : mach_task_self(), &addr, &exSize,
+                            VM_REGION_EXTENDED_INFO,
+                            reinterpret_cast<vm_region_info_t>(&exInfo),
+                            &exInfoCount, &exObjectName);
+        if (kr == KERN_INVALID_ADDRESS) {
+          // Done iterating VM regions.
+          break;
+        } else if (kr != KERN_SUCCESS) {
+          return NS_ERROR_FAILURE;
+        }
+
+        if (exInfo.share_mode == SM_PRIVATE_ALIASED) {
+          privatePages += exInfo.pages_resident;
+        }
+        break;
+      }
       default:
         break;
     }
@@ -522,6 +560,24 @@ static bool InSharedRegion(mach_vm_address_t aAddr, cpu_type_t aType) {
   return NS_OK;
 }
 
+[[nodiscard]] static nsresult PhysicalFootprintAmount(int64_t* aN,
+                                                      mach_port_t aPort = 0) {
+  MOZ_ASSERT(aN);
+
+  // The phys_footprint value (introduced in 10.11) of the TASK_VM_INFO data
+  // matches the value in the 'Memory' column of the Activity Monitor.
+  task_vm_info_data_t task_vm_info;
+  mach_msg_type_number_t count = TASK_VM_INFO_COUNT;
+  kern_return_t kr = task_info(aPort ? aPort : mach_task_self(), TASK_VM_INFO,
+                               (task_info_t)&task_vm_info, &count);
+  if (kr != KERN_SUCCESS) {
+    return NS_ERROR_FAILURE;
+  }
+
+  *aN = task_vm_info.phys_footprint;
+  return NS_OK;
+}
+
 #elif defined(XP_WIN)
 
 #  include <windows.h>
@@ -1009,16 +1065,24 @@ class ResidentUniqueReporter final : public nsIMemoryReporter {
   NS_IMETHOD CollectReports(nsIHandleReportCallback* aHandleReport,
                             nsISupports* aData, bool aAnonymize) override {
     int64_t amount = 0;
+    // clang-format off
     if (NS_SUCCEEDED(ResidentUniqueDistinguishedAmount(&amount))) {
-      // clang-format off
       MOZ_COLLECT_REPORT(
         "resident-unique", KIND_OTHER, UNITS_BYTES, amount,
 "Memory mapped by the process that is present in physical memory and not "
 "shared with any other processes.  This is also known as the process's unique "
 "set size (USS).  This is the amount of RAM we'd expect to be freed if we "
 "closed this process.");
-      // clang-format on
     }
+#ifdef XP_MACOSX
+    if (NS_SUCCEEDED(PhysicalFootprintAmount(&amount))) {
+      MOZ_COLLECT_REPORT(
+        "resident-phys-footprint", KIND_OTHER, UNITS_BYTES, amount,
+"Memory footprint reported by MacOS's task_info API's phys_footprint field. "
+"This matches the memory column in Activity Monitor.");
+    }
+#endif
+    // clang-format on
     return NS_OK;
   }
 };
@@ -1184,16 +1248,19 @@ NS_IMPL_ISUPPORTS(PageFaultsHardReporter, nsIMemoryReporter)
 
 #ifdef HAVE_JEMALLOC_STATS
 
-static size_t HeapOverhead(jemalloc_stats_t* aStats) {
-  return aStats->waste + aStats->bookkeeping + aStats->page_cache +
-         aStats->bin_unused;
+static size_t HeapOverhead(const jemalloc_stats_t& aStats) {
+  return aStats.waste + aStats.bookkeeping + aStats.pages_dirty +
+         aStats.bin_unused;
 }
 
 // This has UNITS_PERCENTAGE, so it is multiplied by 100x *again* on top of the
 // 100x for the percentage.
-static int64_t HeapOverheadFraction(jemalloc_stats_t* aStats) {
+
+// static
+int64_t nsMemoryReporterManager::HeapOverheadFraction(
+    const jemalloc_stats_t& aStats) {
   size_t heapOverhead = HeapOverhead(aStats);
-  size_t heapCommitted = aStats->allocated + heapOverhead;
+  size_t heapCommitted = aStats.allocated + heapOverhead;
   return int64_t(10000 * (heapOverhead / (double)heapCommitted));
 }
 
@@ -1206,12 +1273,14 @@ class JemallocHeapReporter final : public nsIMemoryReporter {
   NS_IMETHOD CollectReports(nsIHandleReportCallback* aHandleReport,
                             nsISupports* aData, bool aAnonymize) override {
     jemalloc_stats_t stats;
-    jemalloc_bin_stats_t bin_stats[JEMALLOC_MAX_STATS_BINS];
-    jemalloc_stats(&stats, bin_stats);
+    const size_t num_bins = jemalloc_stats_num_bins();
+    nsTArray<jemalloc_bin_stats_t> bin_stats(num_bins);
+    bin_stats.SetLength(num_bins);
+    jemalloc_stats(&stats, bin_stats.Elements());
 
     // clang-format off
     MOZ_COLLECT_REPORT(
-      "heap-committed/allocated", KIND_OTHER, UNITS_BYTES, stats.allocated,
+      "heap/committed/allocated", KIND_OTHER, UNITS_BYTES, stats.allocated,
 "Memory mapped by the heap allocator that is currently allocated to the "
 "application.  This may exceed the amount of memory requested by the "
 "application because the allocator regularly rounds up request sizes. (The "
@@ -1219,16 +1288,14 @@ class JemallocHeapReporter final : public nsIMemoryReporter {
 
     MOZ_COLLECT_REPORT(
       "heap-allocated", KIND_OTHER, UNITS_BYTES, stats.allocated,
-"The same as 'heap-committed/allocated'.");
+"The same as 'heap/committed/allocated'.");
 
-    // We mark this and the other heap-overhead reporters as KIND_NONHEAP
+    // We mark this and the other heap/committed/overhead reporters as KIND_NONHEAP
     // because KIND_HEAP memory means "counted in heap-allocated", which
     // this is not.
     for (auto& bin : bin_stats) {
-      if (!bin.size) {
-        continue;
-      }
-      nsPrintfCString path("explicit/heap-overhead/bin-unused/bin-%zu",
+      MOZ_ASSERT(bin.size);
+      nsPrintfCString path("heap/committed/bin-unused/bin-%zu",
           bin.size);
       aHandleReport->Callback(EmptyCString(), path, KIND_NONHEAP, UNITS_BYTES,
         bin.bytes_unused,
@@ -1239,39 +1306,85 @@ class JemallocHeapReporter final : public nsIMemoryReporter {
 
     if (stats.waste > 0) {
       MOZ_COLLECT_REPORT(
-        "explicit/heap-overhead/waste", KIND_NONHEAP, UNITS_BYTES,
+        "heap/committed/waste", KIND_NONHEAP, UNITS_BYTES,
         stats.waste,
 "Committed bytes which do not correspond to an active allocation and which the "
 "allocator is not intentionally keeping alive (i.e., not "
-"'explicit/heap-overhead/{bookkeeping,page-cache,bin-unused}').");
+"'heap/{bookkeeping,unused-pages,bin-unused}').");
     }
 
     MOZ_COLLECT_REPORT(
-      "explicit/heap-overhead/bookkeeping", KIND_NONHEAP, UNITS_BYTES,
+      "heap/committed/bookkeeping", KIND_NONHEAP, UNITS_BYTES,
       stats.bookkeeping,
 "Committed bytes which the heap allocator uses for internal data structures.");
 
     MOZ_COLLECT_REPORT(
-      "explicit/heap-overhead/page-cache", KIND_NONHEAP, UNITS_BYTES,
-      stats.page_cache,
+      "heap/committed/unused-pages/dirty", KIND_NONHEAP, UNITS_BYTES,
+      stats.pages_dirty,
 "Memory which the allocator could return to the operating system, but hasn't. "
 "The allocator keeps this memory around as an optimization, so it doesn't "
 "have to ask the OS the next time it needs to fulfill a request. This value "
 "is typically not larger than a few megabytes.");
 
     MOZ_COLLECT_REPORT(
-      "heap-committed/overhead", KIND_OTHER, UNITS_BYTES,
-      HeapOverhead(&stats),
-"The sum of 'explicit/heap-overhead/*'.");
-
+      "heap/decommitted/unused-pages/fresh", KIND_OTHER, UNITS_BYTES, stats.pages_fresh,
+"Amount of memory currently mapped but has never been used.");
+    // A duplicate entry in the decommitted part of the tree.
     MOZ_COLLECT_REPORT(
-      "heap-mapped", KIND_OTHER, UNITS_BYTES, stats.mapped,
-"Amount of memory currently mapped. Includes memory that is uncommitted, i.e. "
-"neither in physical memory nor paged to disk.");
+      "decommitted/heap/unused-pages/fresh", KIND_OTHER, UNITS_BYTES, stats.pages_fresh,
+"Amount of memory currently mapped but has never been used.");
 
+// On MacOS madvised memory is still counted in the resident set until the OS
+// actually decommits it.
+#ifdef XP_MACOSX
+#define MADVISED_GROUP "committed"
+#else
+#define MADVISED_GROUP "decommitted"
+#endif
+    MOZ_COLLECT_REPORT(
+      "heap/" MADVISED_GROUP "/unused-pages/madvised", KIND_OTHER, UNITS_BYTES,
+      stats.pages_madvised,
+"Amount of memory currently mapped, not used and that the OS should remove "
+"from the application's resident set.");
+    // A duplicate entry in the decommitted part of the tree.
+    MOZ_COLLECT_REPORT(
+      "decommitted/heap/unused-pages/madvised", KIND_OTHER, UNITS_BYTES, stats.pages_madvised,
+"Amount of memory currently mapped, not used and that the OS should remove "
+"from the application's resident set.");
+
+    {
+      size_t decommitted = stats.mapped - stats.allocated - stats.waste - stats.pages_dirty - stats.pages_fresh - stats.bookkeeping - stats.bin_unused;
+      MOZ_COLLECT_REPORT(
+        "heap/decommitted/unmapped", KIND_OTHER, UNITS_BYTES, decommitted,
+  "Amount of memory currently mapped but not committed, "
+  "neither in physical memory nor paged to disk.");
+      MOZ_COLLECT_REPORT(
+        "decommitted/heap/decommitted", KIND_OTHER, UNITS_BYTES, decommitted,
+  "Amount of memory currently mapped but not committed, "
+  "neither in physical memory nor paged to disk.");
+    }
     MOZ_COLLECT_REPORT(
       "heap-chunksize", KIND_OTHER, UNITS_BYTES, stats.chunksize,
       "Size of chunks.");
+
+#ifdef MOZ_PHC
+    mozilla::phc::MemoryUsage usage;
+    mozilla::phc::PHCMemoryUsage(usage);
+
+    MOZ_COLLECT_REPORT(
+      "explicit/phc/metadata", KIND_NONHEAP, UNITS_BYTES,
+      usage.mMetadataBytes,
+"Memory used by PHC to store stacks and other metadata for each allocation");
+    MOZ_COLLECT_REPORT(
+      "explicit/phc/fragmentation", KIND_NONHEAP, UNITS_BYTES,
+      usage.mFragmentationBytes,
+"The amount of memory lost due to rounding up allocations to the next page "
+"size. "
+"This is also known as 'internal fragmentation'. "
+"Note that all allocators have some internal fragmentation, there may still "
+"be some internal fragmentation without PHC.");
+#endif
+
     // clang-format on
 
     return NS_OK;
@@ -1341,78 +1454,87 @@ class ThreadsReporter final : public nsIMemoryReporter {
     size_t wrapperSizes = 0;
     size_t threadCount = 0;
 
-    for (auto* thread : nsThread::Enumerate()) {
-      threadCount++;
-      eventQueueSizes += thread->SizeOfEventQueues(MallocSizeOf);
-      wrapperSizes += thread->ShallowSizeOfIncludingThis(MallocSizeOf);
+    {
+      nsThreadManager& tm = nsThreadManager::get();
+      OffTheBooksMutexAutoLock lock(tm.ThreadListMutex());
+      for (auto* thread : tm.ThreadList()) {
+        threadCount++;
+        eventQueueSizes += thread->SizeOfEventQueues(MallocSizeOf);
+        wrapperSizes += thread->ShallowSizeOfIncludingThis(MallocSizeOf);
 
-      if (!thread->StackBase()) {
-        continue;
-      }
+        if (!thread->StackBase()) {
+          continue;
+        }
 
 #if defined(XP_LINUX)
-      int idx = mappings.BinaryIndexOf(thread->StackBase());
-      if (idx < 0) {
-        continue;
-      }
-      // Referenced() is the combined size of all pages in the region which have
-      // ever been touched, and are therefore consuming memory. For stack
-      // regions, these pages are guaranteed to be un-shared unless we fork
-      // after creating threads (which we don't).
-      size_t privateSize = mappings[idx].Referenced();
+        int idx = mappings.BinaryIndexOf(thread->StackBase());
+        if (idx < 0) {
+          continue;
+        }
+        // Referenced() is the combined size of all pages in the region which
+        // have ever been touched, and are therefore consuming memory. For stack
+        // regions, these pages are guaranteed to be un-shared unless we fork
+        // after creating threads (which we don't).
+        size_t privateSize = mappings[idx].Referenced();
 
-      // On Linux, we have to be very careful matching memory regions to thread
-      // stacks.
-      //
-      // To begin with, the kernel only reports VM stats for regions of all
-      // adjacent pages with the same flags, protection, and backing file.
-      // There's no way to get finer-grained usage information for a subset of
-      // those pages.
-      //
-      // Stack segments always have a guard page at the bottom of the stack
-      // (assuming we only support stacks that grow down), so there's no danger
-      // of them being merged with other stack regions. At the top, there's no
-      // protection page, and no way to allocate one without using pthreads
-      // directly and allocating our own stacks. So we get around the problem by
-      // adding an extra VM flag (NOHUGEPAGES) to our stack region, which we
-      // don't expect to be set on any heap regions. But this is not fool-proof.
-      //
-      // A second kink is that different C libraries (and different versions
-      // thereof) report stack base locations and sizes differently with regard
-      // to the guard page. For the libraries that include the guard page in the
-      // stack size base pointer, we need to adjust those values to compensate.
-      // But it's possible that our logic will get out of sync with library
-      // changes, or someone will compile with an unexpected library.
-      //
-      //
-      // The upshot of all of this is that there may be configurations that our
-      // special cases don't cover. And if there are, we want to know about it.
-      // So assert that total size of the memory region we're reporting actually
-      // matches the allocated size of the thread stack.
+        // On Linux, we have to be very careful matching memory regions to
+        // thread stacks.
+        //
+        // To begin with, the kernel only reports VM stats for regions of all
+        // adjacent pages with the same flags, protection, and backing file.
+        // There's no way to get finer-grained usage information for a subset of
+        // those pages.
+        //
+        // Stack segments always have a guard page at the bottom of the stack
+        // (assuming we only support stacks that grow down), so there's no
+        // danger of them being merged with other stack regions. At the top,
+        // there's no protection page, and no way to allocate one without using
+        // pthreads directly and allocating our own stacks. So we get around the
+        // problem by adding an extra VM flag (NOHUGEPAGES) to our stack region,
+        // which we don't expect to be set on any heap regions. But this is not
+        // fool-proof.
+        //
+        // A second kink is that different C libraries (and different versions
+        // thereof) report stack base locations and sizes differently with
+        // regard to the guard page. For the libraries that include the guard
+        // page in the stack size base pointer, we need to adjust those values
+        // to compensate. But it's possible that our logic will get out of sync
+        // with library changes, or someone will compile with an unexpected
+        // library.
+        //
+        //
+        // The upshot of all of this is that there may be configurations that
+        // our special cases don't cover. And if there are, we want to know
+        // about it. So assert that total size of the memory region we're
+        // reporting actually matches the allocated size of the thread stack.
 #  ifndef ANDROID
-      MOZ_ASSERT(mappings[idx].Size() == thread->StackSize(),
-                 "Mapping region size doesn't match stack allocation size");
+        MOZ_ASSERT(mappings[idx].Size() == thread->StackSize(),
+                   "Mapping region size doesn't match stack allocation size");
 #  endif
 #elif defined(XP_WIN)
-      auto memInfo = MemoryInfo::Get(thread->StackBase(), thread->StackSize());
-      size_t privateSize = memInfo.Committed();
+        auto memInfo =
+            MemoryInfo::Get(thread->StackBase(), thread->StackSize());
+        size_t privateSize = memInfo.Committed();
 #else
-      size_t privateSize = thread->StackSize();
-      MOZ_ASSERT_UNREACHABLE(
-          "Shouldn't have stack base pointer on this "
-          "platform");
+        size_t privateSize = thread->StackSize();
+        MOZ_ASSERT_UNREACHABLE(
+            "Shouldn't have stack base pointer on this "
+            "platform");
 #endif
 
-      threads.AppendElement(ThreadData{
-          nsCString(PR_GetThreadName(thread->GetPRThread())),
-          thread->ThreadId(),
-          // On Linux, it's possible (but unlikely) that our stack region will
-          // have been merged with adjacent heap regions, in which case we'll
-          // get combined size information for both. So we take the minimum of
-          // the reported private size and the requested stack size to avoid the
-          // possible of majorly over-reporting in that case.
-          std::min(privateSize, thread->StackSize()),
-      });
+        nsCString threadName;
+        thread->GetThreadName(threadName);
+        threads.AppendElement(ThreadData{
+            std::move(threadName),
+            thread->ThreadId(),
+            // On Linux, it's possible (but unlikely) that our stack region will
+            // have been merged with adjacent heap regions, in which case we'll
+            // get combined size information for both. So we take the minimum of
+            // the reported private size and the requested stack size to avoid
+            // the possible of majorly over-reporting in that case.
+            std::min(privateSize, thread->StackSize()),
+        });
+      }
     }
 
     for (auto& thread : threads) {
@@ -1548,6 +1670,35 @@ NS_IMPL_ISUPPORTS(DMDReporter, nsIMemoryReporter)
 
 #endif  // MOZ_DMD
 
+#ifdef MOZ_WIDGET_ANDROID
+class AndroidMemoryReporter final : public nsIMemoryReporter {
+ public:
+  NS_DECL_ISUPPORTS
+
+  AndroidMemoryReporter() = default;
+
+  NS_IMETHOD
+  CollectReports(nsIHandleReportCallback* aHandleReport, nsISupports* aData,
+                 bool aAnonymize) override {
+    if (!jni::IsAvailable() || jni::GetAPIVersion() < 23) {
+      return NS_OK;
+    }
+
+    int32_t heap = java::GeckoAppShell::GetMemoryUsage("summary.java-heap"_ns);
+    if (heap > 0) {
+      MOZ_COLLECT_REPORT("java-heap", KIND_OTHER, UNITS_BYTES, heap * 1024,
+                         "The private Java Heap usage");
+    }
+    return NS_OK;
+  }
+
+ private:
+  ~AndroidMemoryReporter() = default;
+};
+
+NS_IMPL_ISUPPORTS(AndroidMemoryReporter, nsIMemoryReporter)
+#endif
+
 /**
  ** nsMemoryReporterManager implementation
  **/
@@ -1581,77 +1732,93 @@ nsMemoryReporterManager::Init() {
   }
   isInited = true;
 
+  {
+    // Add a series of low-level reporters meant to be executed in order and
+    // before any other reporters. These reporters are never released until
+    // the manager dies (at process shutdown). Note that this currently only
+    // works for reporters expecting to be executed sync.
+    //
+    // Note that we explicitly handle our self-reporting inside
+    // GetReportsForThisProcessExtended, such that we do not need to register
+    // ourself to any array/table here.
+
+    mozilla::MutexAutoLock autoLock(mMutex);
+
 #ifdef HAVE_JEMALLOC_STATS
-  RegisterStrongReporter(new JemallocHeapReporter());
+    mStrongEternalReporters->AppendElement(new JemallocHeapReporter());
 #endif
 
 #ifdef HAVE_VSIZE_AND_RESIDENT_REPORTERS
-  RegisterStrongReporter(new VsizeReporter());
-  RegisterStrongReporter(new ResidentReporter());
+    mStrongEternalReporters->AppendElement(new VsizeReporter());
+    mStrongEternalReporters->AppendElement(new ResidentReporter());
 #endif
 
 #ifdef HAVE_VSIZE_MAX_CONTIGUOUS_REPORTER
-  RegisterStrongReporter(new VsizeMaxContiguousReporter());
+    mStrongEternalReporters->AppendElement(new VsizeMaxContiguousReporter());
 #endif
 
 #ifdef HAVE_RESIDENT_PEAK_REPORTER
-  RegisterStrongReporter(new ResidentPeakReporter());
+    mStrongEternalReporters->AppendElement(new ResidentPeakReporter());
 #endif
 
 #ifdef HAVE_RESIDENT_UNIQUE_REPORTER
-  RegisterStrongReporter(new ResidentUniqueReporter());
+    mStrongEternalReporters->AppendElement(new ResidentUniqueReporter());
 #endif
 
 #ifdef HAVE_PAGE_FAULT_REPORTERS
-  RegisterStrongReporter(new PageFaultsSoftReporter());
-  RegisterStrongReporter(new PageFaultsHardReporter());
+    mStrongEternalReporters->AppendElement(new PageFaultsSoftReporter());
+    mStrongEternalReporters->AppendElement(new PageFaultsHardReporter());
 #endif
 
 #ifdef HAVE_PRIVATE_REPORTER
-  RegisterStrongReporter(new PrivateReporter());
+    mStrongEternalReporters->AppendElement(new PrivateReporter());
 #endif
 
 #ifdef HAVE_SYSTEM_HEAP_REPORTER
-  RegisterStrongReporter(new SystemHeapReporter());
+    mStrongEternalReporters->AppendElement(new SystemHeapReporter());
 #endif
 
-  RegisterStrongReporter(new AtomTablesReporter());
+    mStrongEternalReporters->AppendElement(new AtomTablesReporter());
 
-  RegisterStrongReporter(new ThreadsReporter());
+    mStrongEternalReporters->AppendElement(new ThreadsReporter());
 
 #ifdef DEBUG
-  RegisterStrongReporter(new DeadlockDetectorReporter());
+    mStrongEternalReporters->AppendElement(new DeadlockDetectorReporter());
 #endif
 
 #ifdef MOZ_GECKO_PROFILER
-  // We have to register this here rather than in profiler_init() because
-  // profiler_init() runs prior to nsMemoryReporterManager's creation.
-  RegisterStrongReporter(new GeckoProfilerReporter());
+    // We have to register this here rather than in profiler_init() because
+    // profiler_init() runs prior to nsMemoryReporterManager's creation.
+    mStrongEternalReporters->AppendElement(new GeckoProfilerReporter());
 #endif
 
 #ifdef MOZ_DMD
-  RegisterStrongReporter(new mozilla::dmd::DMDReporter());
+    mStrongEternalReporters->AppendElement(new mozilla::dmd::DMDReporter());
 #endif
 
 #ifdef XP_WIN
-  RegisterStrongReporter(new WindowsAddressSpaceReporter());
+    mStrongEternalReporters->AppendElement(new WindowsAddressSpaceReporter());
 #endif
 
+#ifdef MOZ_WIDGET_ANDROID
+    mStrongEternalReporters->AppendElement(new AndroidMemoryReporter());
+#endif
+  }  // autoLock(mMutex);
+
 #ifdef XP_UNIX
   nsMemoryInfoDumper::Initialize();
 #endif
 
-  // Report our own memory usage as well.
-  RegisterWeakReporter(this);
-
   return NS_OK;
 }
 
 nsMemoryReporterManager::nsMemoryReporterManager()
     : mMutex("nsMemoryReporterManager::mMutex"),
       mIsRegistrationBlocked(false),
+      mStrongEternalReporters(new StrongReportersArray()),
       mStrongReporters(new StrongReportersTable()),
       mWeakReporters(new WeakReportersTable()),
+      mSavedStrongEternalReporters(nullptr),
       mSavedStrongReporters(nullptr),
       mSavedWeakReporters(nullptr),
       mNextGeneration(1),
@@ -1665,8 +1832,8 @@ nsMemoryReporterManager::nsMemoryReporterManager()
 }
 
 nsMemoryReporterManager::~nsMemoryReporterManager() {
-  delete mStrongReporters;
-  delete mWeakReporters;
+  NS_ASSERTION(!mSavedStrongEternalReporters,
+               "failed to restore eternal reporters");
   NS_ASSERTION(!mSavedStrongReporters, "failed to restore strong reporters");
   NS_ASSERTION(!mSavedWeakReporters, "failed to restore weak reporters");
 }
@@ -1675,8 +1842,15 @@ NS_IMETHODIMP
 nsMemoryReporterManager::CollectReports(nsIHandleReportCallback* aHandleReport,
                                         nsISupports* aData, bool aAnonymize) {
   size_t n = MallocSizeOf(this);
-  n += mStrongReporters->ShallowSizeOfIncludingThis(MallocSizeOf);
-  n += mWeakReporters->ShallowSizeOfIncludingThis(MallocSizeOf);
+  {
+    mozilla::MutexAutoLock autoLock(mMutex);
+    n += mStrongEternalReporters->ShallowSizeOfIncludingThis(MallocSizeOf);
+    n += mStrongReporters->ShallowSizeOfIncludingThis(MallocSizeOf);
+    n += mWeakReporters->ShallowSizeOfIncludingThis(MallocSizeOf);
+    // Note that we do not include the mSaved<X>Reporters here, as during
+    // normal operations they are always nullptr and during testing we want to
+    // hide the saved variants entirely.
+  }
 
   MOZ_COLLECT_REPORT("explicit/memory-reporter-manager", KIND_HEAP, UNITS_BYTES,
                      n, "Memory used by the memory reporter infrastructure.");
@@ -1748,6 +1922,7 @@ nsMemoryReporterManager::GetReportsExtended(
   return rv;
 }
 
+// MainThread only
 nsresult nsMemoryReporterManager::StartGettingReports() {
   PendingProcessesState* s = mPendingProcessesState;
   nsresult rv;
@@ -1801,13 +1976,26 @@ nsresult nsMemoryReporterManager::StartGettingReports() {
     }
   }
 
-  if (!mIsRegistrationBlocked && net::gIOService) {
+  if (!IsRegistrationBlocked() && net::gIOService) {
     if (RefPtr<MemoryReportingProcess> proc =
             net::gIOService->GetSocketProcessMemoryReporter()) {
       s->mChildrenPending.AppendElement(proc.forget());
     }
   }
 
+  if (!IsRegistrationBlocked()) {
+    if (RefPtr<UtilityProcessManager> utility =
+            UtilityProcessManager::GetIfExists()) {
+      for (RefPtr<UtilityProcessParent>& parent :
+           utility->GetAllProcessesProcessParent()) {
+        if (RefPtr<MemoryReportingProcess> proc =
+                utility->GetProcessMemoryReporter(parent)) {
+          s->mChildrenPending.AppendElement(proc.forget());
+        }
+      }
+    }
+  }
+
   if (!s->mChildrenPending.IsEmpty()) {
     nsCOMPtr<nsITimer> timer;
     rv = NS_NewTimerWithFuncCallback(
@@ -1884,21 +2072,37 @@ nsMemoryReporterManager::GetReportsForThisProcessExtended(
   {
     mozilla::MutexAutoLock autoLock(mMutex);
 
-    for (auto iter = mStrongReporters->Iter(); !iter.Done(); iter.Next()) {
-      DispatchReporter(iter.Key(), iter.Data(), aHandleReport,
-                       aHandleReportData, aAnonymize);
+    // We process our own, most sensible reporters before anyone else to avoid
+    // them measuring changes caused by other reporters' dynamic structures.
+    // Note that all eternal reporters need to be sync, too.
+    for (const auto& entry : *mStrongEternalReporters) {
+      DispatchReporter(entry, false, aHandleReport, aHandleReportData,
+                       aAnonymize);
+    }
+    // Process our self-reporting (not in any array/table). Note that when
+    // we test, we expect to execute only reporters in the swapped-in tables.
+    if (!mIsRegistrationBlocked) {
+      DispatchReporter(this, false, aHandleReport, aHandleReportData,
+                       aAnonymize);
     }
 
-    for (auto iter = mWeakReporters->Iter(); !iter.Done(); iter.Next()) {
-      nsCOMPtr<nsIMemoryReporter> reporter = iter.Key();
-      DispatchReporter(reporter, iter.Data(), aHandleReport, aHandleReportData,
-                       aAnonymize);
+    // Now process additional reporters. Note that these are executed in an
+    // unforeseeable order (due to the hashtables being keyed on pointers).
+    for (const auto& entry : *mStrongReporters) {
+      DispatchReporter(entry.GetKey(), entry.GetData(), aHandleReport,
+                       aHandleReportData, aAnonymize);
+    }
+    for (const auto& entry : *mWeakReporters) {
+      nsCOMPtr<nsIMemoryReporter> reporter = entry.GetKey();
+      DispatchReporter(reporter, entry.GetData(), aHandleReport,
+                       aHandleReportData, aAnonymize);
     }
   }
 
   return NS_OK;
 }
 
+// MainThread only
 NS_IMETHODIMP
 nsMemoryReporterManager::EndReport() {
   if (--mPendingReportersState->mReportsPending == 0) {
@@ -2101,7 +2305,6 @@ nsMemoryReporterManager::PendingProcessesState::PendingProcessesState(
     : mGeneration(aGeneration),
       mAnonymize(aAnonymize),
       mMinimize(aMinimize),
-      mChildrenPending(),
       mNumProcessesRunning(1),  // reporting starts with the parent
       mNumProcessesCompleted(0),
       mConcurrencyLimit(aConcurrencyLimit),
@@ -2147,7 +2350,7 @@ nsresult nsMemoryReporterManager::RegisterReporterHelper(
   //
   if (aStrong) {
     nsCOMPtr<nsIMemoryReporter> kungFuDeathGrip = aReporter;
-    mStrongReporters->Put(aReporter, aIsAsync);
+    mStrongReporters->InsertOrUpdate(aReporter, aIsAsync);
     CrashIfRefcountIsZero(aReporter);
   } else {
     CrashIfRefcountIsZero(aReporter);
@@ -2160,7 +2363,7 @@ nsresult nsMemoryReporterManager::RegisterReporterHelper(
       // CollectReports().
       return NS_ERROR_XPC_BAD_CONVERT_JS;
     }
-    mWeakReporters->Put(aReporter, aIsAsync);
+    mWeakReporters->InsertOrUpdate(aReporter, aIsAsync);
   }
 
   return NS_OK;
@@ -2263,12 +2466,15 @@ nsMemoryReporterManager::BlockRegistrationAndHideExistingReporters() {
   mIsRegistrationBlocked = true;
 
   // Hide the existing reporters, saving them for later restoration.
+  MOZ_ASSERT(!mSavedStrongEternalReporters);
   MOZ_ASSERT(!mSavedStrongReporters);
   MOZ_ASSERT(!mSavedWeakReporters);
-  mSavedStrongReporters = mStrongReporters;
-  mSavedWeakReporters = mWeakReporters;
-  mStrongReporters = new StrongReportersTable();
-  mWeakReporters = new WeakReportersTable();
+  mSavedStrongEternalReporters.swap(mStrongEternalReporters);
+  mSavedStrongReporters.swap(mStrongReporters);
+  mSavedWeakReporters.swap(mWeakReporters);
+  mStrongEternalReporters.reset(new StrongReportersArray());
+  mStrongReporters.reset(new StrongReportersTable());
+  mWeakReporters.reset(new WeakReportersTable());
 
   return NS_OK;
 }
@@ -2282,12 +2488,9 @@ nsMemoryReporterManager::UnblockRegistrationAndRestoreOriginalReporters() {
   }
 
   // Banish the current reporters, and restore the hidden ones.
-  delete mStrongReporters;
-  delete mWeakReporters;
-  mStrongReporters = mSavedStrongReporters;
-  mWeakReporters = mSavedWeakReporters;
-  mSavedStrongReporters = nullptr;
-  mSavedWeakReporters = nullptr;
+  mStrongEternalReporters = std::move(mSavedStrongEternalReporters);
+  mStrongReporters = std::move(mSavedStrongReporters);
+  mWeakReporters = std::move(mSavedWeakReporters);
 
   mIsRegistrationBlocked = false;
   return NS_OK;
@@ -2377,6 +2580,16 @@ nsMemoryReporterManager::GetResidentUnique(int64_t* aAmount) {
 #endif
 }
 
+#ifdef XP_MACOSX
+/*static*/
+int64_t nsMemoryReporterManager::PhysicalFootprint(mach_port_t aPort) {
+  int64_t amount = 0;
+  nsresult rv = PhysicalFootprintAmount(&amount, aPort);
+  NS_ENSURE_SUCCESS(rv, 0);
+  return amount;
+}
+#endif
+
 typedef
 #ifdef XP_WIN
     HANDLE
@@ -2415,12 +2628,19 @@ int64_t nsMemoryReporterManager::ResidentUnique(ResidentUniqueArg) {
 
 #endif  // XP_{WIN, MACOSX, LINUX, *}
 
+#ifdef HAVE_JEMALLOC_STATS
+// static
+size_t nsMemoryReporterManager::HeapAllocated(const jemalloc_stats_t& aStats) {
+  return aStats.allocated;
+}
+#endif
+
 NS_IMETHODIMP
 nsMemoryReporterManager::GetHeapAllocated(int64_t* aAmount) {
 #ifdef HAVE_JEMALLOC_STATS
   jemalloc_stats_t stats;
   jemalloc_stats(&stats);
-  *aAmount = stats.allocated;
+  *aAmount = HeapAllocated(stats);
   return NS_OK;
 #else
   *aAmount = 0;
@@ -2434,7 +2654,7 @@ nsMemoryReporterManager::GetHeapOverheadFraction(int64_t* aAmount) {
 #ifdef HAVE_JEMALLOC_STATS
   jemalloc_stats_t stats;
   jemalloc_stats(&stats);
-  *aAmount = HeapOverheadFraction(&stats);
+  *aAmount = HeapOverheadFraction(stats);
   return NS_OK;
 #else
   *aAmount = 0;
@@ -2495,16 +2715,6 @@ nsMemoryReporterManager::GetStorageSQLite(int64_t* aAmount) {
   return GetInfallibleAmount(mAmountFns.mStorageSQLite, aAmount);
 }
 
-NS_IMETHODIMP
-nsMemoryReporterManager::GetLowMemoryEventsVirtual(int64_t* aAmount) {
-  return GetInfallibleAmount(mAmountFns.mLowMemoryEventsVirtual, aAmount);
-}
-
-NS_IMETHODIMP
-nsMemoryReporterManager::GetLowMemoryEventsCommitSpace(int64_t* aAmount) {
-  return GetInfallibleAmount(mAmountFns.mLowMemoryEventsCommitSpace, aAmount);
-}
-
 NS_IMETHODIMP
 nsMemoryReporterManager::GetLowMemoryEventsPhysical(int64_t* aAmount) {
   return GetInfallibleAmount(mAmountFns.mLowMemoryEventsPhysical, aAmount);
@@ -2581,6 +2791,17 @@ class MinimizeMemoryUsageRunnable : public Runnable {
     }
 
     if (mRemainingIters == 0) {
+#ifdef MOZ_MEMORY
+      // The nsMemoryPressureWatcher observes "memory-pressure" and dispatches
+      // a nsJemallocFreeDirtyPagesRunnable that will be run after the other
+      // observers have been called. If those observers did synchronously
+      // whatever they want to do without dispatching other runnables to the
+      // main thread, this is guaranteeing that the last thing to run before
+      // coming here was that nsJemallocFreeDirtyPagesRunnable. But for
+      // paranoia we do another purge here which ideally is a no-op.
+      jemalloc_free_dirty_pages();
+#endif
+
       os->NotifyObservers(nullptr, "after-minimize-memory-usage",
                           u"MinimizeMemoryUsageRunnable");
       if (mCallback) {
@@ -2749,8 +2970,6 @@ DEFINE_UNREGISTER_DISTINGUISHED_AMOUNT(ImagesContentUsedUncompressed)
 DEFINE_REGISTER_DISTINGUISHED_AMOUNT(Infallible, StorageSQLite)
 DEFINE_UNREGISTER_DISTINGUISHED_AMOUNT(StorageSQLite)
 
-DEFINE_REGISTER_DISTINGUISHED_AMOUNT(Infallible, LowMemoryEventsVirtual)
-DEFINE_REGISTER_DISTINGUISHED_AMOUNT(Infallible, LowMemoryEventsCommitSpace)
 DEFINE_REGISTER_DISTINGUISHED_AMOUNT(Infallible, LowMemoryEventsPhysical)
 
 DEFINE_REGISTER_DISTINGUISHED_AMOUNT(Infallible, GhostWindows)
