@@ -8,17 +8,31 @@
 #  error "Don't include this file directly"
 #endif
 
+// NOTE: We don't use MOZ_COUNT_CTOR/MOZ_COUNT_DTOR to perform leak checking of
+// nsTArray_base objects intentionally for the following reasons:
+// * The leak logging isn't as useful as other types of logging, as
+//   nsTArray_base is frequently relocated without invoking a constructor, such
+//   as when stored within another nsTArray. This means that
+//   XPCOM_MEM_LOG_CLASSES cannot be used to identify specific leaks of nsTArray
+//   objects.
+// * The nsTArray type is layout compatible with the ThinVec crate with the
+//   correct flags, and ThinVec does not currently perform leak logging.
+//   This means that if a large number of arrays are transferred between Rust
+//   and C++ code using ThinVec, for example within another ThinVec, they
+//   will not be logged correctly and might appear as e.g. negative leaks.
+// * Leaks which have been found thanks to the leak logging added by this
+//   type have often not been significant, and/or have needed to be
+//   circumvented using some other mechanism. Most leaks found with this type
+//   in them also include other types which will continue to be tracked.
+
 template <class Alloc, class RelocationStrategy>
-nsTArray_base<Alloc, RelocationStrategy>::nsTArray_base() : mHdr(EmptyHdr()) {
-  MOZ_COUNT_CTOR(nsTArray_base);
-}
+nsTArray_base<Alloc, RelocationStrategy>::nsTArray_base() : mHdr(EmptyHdr()) {}
 
 template <class Alloc, class RelocationStrategy>
 nsTArray_base<Alloc, RelocationStrategy>::~nsTArray_base() {
   if (!HasEmptyHeader() && !UsesAutoArrayBuffer()) {
     Alloc::Free(mHdr);
   }
-  MOZ_COUNT_DTOR(nsTArray_base);
 }
 
 template <class Alloc, class RelocationStrategy>
@@ -26,7 +40,6 @@ nsTArray_base<Alloc, RelocationStrategy>::nsTArray_base(const nsTArray_base&)
     : mHdr(EmptyHdr()) {
   // Actual copying happens through nsTArray_CopyEnabler, we just need to do the
   // initialization of mHdr.
-  MOZ_COUNT_CTOR(nsTArray_base);
 }
 
 template <class Alloc, class RelocationStrategy>
@@ -52,7 +65,7 @@ nsTArray_base<Alloc, RelocationStrategy>::GetAutoArrayBufferUnsafe(
   // pointer to take into account the extra alignment in the auto array.
 
   static_assert(
-      sizeof(void*) != 4 || (MOZ_ALIGNOF(mozilla::AlignedElem<8>) == 8 &&
+      sizeof(void*) != 4 || (alignof(mozilla::AlignedElem<8>) == 8 &&
                              sizeof(AutoTArray<mozilla::AlignedElem<8>, 1>) ==
                                  sizeof(void*) + sizeof(nsTArrayHeader) + 4 +
                                      sizeof(mozilla::AlignedElem<8>)),
@@ -75,7 +88,7 @@ bool nsTArray_base<Alloc, RelocationStrategy>::UsesAutoArrayBuffer() const {
 
   // This is nuts.  If we were sane, we'd pass aElemAlign as a parameter to
   // this function.  Unfortunately this function is called in nsTArray_base's
-  // destructor, at which point we don't know elem_type's alignment.
+  // destructor, at which point we don't know value_type's alignment.
   //
   // We'll fall on our face and return true when we should say false if
   //
@@ -100,7 +113,7 @@ bool nsTArray_base<Alloc, RelocationStrategy>::UsesAutoArrayBuffer() const {
   //
   // Note that this means that we can't store elements with alignment 16 in an
   // nsTArray, because GetAutoArrayBuffer(16) could lie outside the memory
-  // owned by this AutoTArray.  We statically assert that elem_type's
+  // owned by this AutoTArray.  We statically assert that value_type's
   // alignment is 8 bytes or less in AutoTArray.
 
   static_assert(sizeof(nsTArrayHeader) > 4, "see comment above");
@@ -138,12 +151,10 @@ nsTArray_base<Alloc, RelocationStrategy>::ExtendCapacity(size_type aLength,
 template <class Alloc, class RelocationStrategy>
 template <typename ActualAlloc>
 typename ActualAlloc::ResultTypeProxy
-nsTArray_base<Alloc, RelocationStrategy>::EnsureCapacity(size_type aCapacity,
-                                                         size_type aElemSize) {
-  // This should be the most common case so test this first
-  if (aCapacity <= mHdr->mCapacity) {
-    return ActualAlloc::SuccessResult();
-  }
+nsTArray_base<Alloc, RelocationStrategy>::EnsureCapacityImpl(
+    size_type aCapacity, size_type aElemSize) {
+  MOZ_ASSERT(aCapacity > mHdr->mCapacity,
+             "Should have been checked by caller (EnsureCapacity)");
 
   // If the requested memory allocation exceeds size_type(-1)/2, then
   // our doubling algorithm may not be able to allocate it.
@@ -399,7 +410,7 @@ nsTArray_base<Alloc, RelocationStrategy>::InsertSlotsAt(index_type aIndex,
                                                         size_type aElemSize,
                                                         size_t aElemAlign) {
   if (MOZ_UNLIKELY(aIndex > Length())) {
-    InvalidArrayIndex_CRASH(aIndex, Length());
+    mozilla::detail::InvalidArrayIndex_CRASH(aIndex, Length());
   }
 
   if (!ActualAlloc::Successful(
